/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef FIRM_AVTREE
#define FIRM_AVTREE

#include <iostream>
using std::ostream;
using std::cout;
using std::endl;

#include <vector>
using std::vector;

#include <string>
using std::string;

class Item;
class Attribute;
class FITree;

/**
 * An attribute-value tree is a regular attribute-value dataset.
 * An attribute-value tree can be formed from for example data given below:
 *
 * BASE    IDLE    VOLUME  ACTIVE_APPLICATION
 * 2       idle    2       none
 * 1       idle    2       none
 * 2       active  7       phonebook
 * 2       active  7       phonebook
 * 3       idle    2       none
 *
 * Or from a transaction dataset like:
 *
 * beer	sausages	diapers
 * beer diapers
 * diapers	milk	ham
 * beer	milk	ham
 * diapers	ham
 *
 * @author Juho Muhonen
 * @modification 25.10.2005 commented all
 */
class AVTree {
private:
	/**
	 * The attributes (the column names).
	 */
	vector<Attribute*> attributes;
	
	/**
	 * The number of transactions (lines)
	 */
	int transactions;
public:
	/**
	 * Create an empty attribute-value tree.
	 */
	AVTree();

	/**
	 * Create an attribute-value tree from a text file.
	 */
	AVTree(string, char, char);
	
	/**
	 * Create an attribute-value tree from a text file. With possibility
	 * of a transaction dataset conversion.
	 */
	AVTree(string, char, char, bool);
	
	/**
	 * The destructor.
	 */
	~AVTree();

	/**
	 * Build an attribute-value tree from a text file.
	 */
	void buildTree(string, char, char);

	/**
	 * Build an attribute-value tree from a text file.
	 */
	void buildTree(const char*, char, char);

	/**
	 * Build an attribute-value tree from a text file containing a
	 * transaction dataset.
	 */
	void buildTreeTransaction(string, char, char);

	/**
	 * Build an attribute-value tree from a text file containing a
	 * transaction dataset.
	 */
	void buildTreeTransaction(const char*, char, char);
	
	/**
	 * Get the frequent items in this dataset.
	 */
	FITree* frequentItems(int);
	
	/**
	 * Get the number of transactions in this dataset.
	 */
	int getTransactions();

	/**
	 * Output this tree to a stream.
	 */
	friend ostream& operator<<(ostream&, const AVTree&);
};

ostream& operator<<(ostream&, const AVTree&);
#endif
