/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "ContextAverageDurationConstraint.h"
#include "ContextTree.h"
#include "FITree.h"
#include "AssociationRule.h"
#include "Cover.h"

/**
 * Create a context frequency constraint with minimum frequency of 0.
 */
ContextAverageDurationConstraint::ContextAverageDurationConstraint() {
	last_duration = 0;
	min_duration = 0;
}

/**
 * Create a context frequency constraint with a minimum frequency.
 * @param d The minimum average duration.
 */
ContextAverageDurationConstraint::ContextAverageDurationConstraint(double d) {
	last_duration = 0;
	min_duration = d;
}

/**
 * Set the minimum average duration of this constraint.
 * @param d The minimum average duration.
 */
void ContextAverageDurationConstraint::setMinDuration(double d) {
	min_duration = d;
}

/**
 * Returns the result of last evaluation.
 * @return The average duration of the pattern evaluated last.
 */
double ContextAverageDurationConstraint::lastResult() {
	return last_duration;
}

/**
 * Return the average duration of an itemset.
 * @param fset The itemset.
 * @return The average duration of the itemset.
 */
double ContextAverageDurationConstraint::value(FITree* fset) {
	Cover* cover = fset->getCover();

	vector<int> tids;
	for (int i = 0; i < cover->size; i++) {
		tids.push_back(cover->tid_table[i]);
	}
	sort(tids.begin(), tids.end());

	int context_changes = 0;
	for (int i = 0; i < tids.size(); i++) {
		if (i == 0) {
			context_changes++;
		} else if (tids[i] - 1 != tids[i-1]) {
			context_changes++;
		}
	}

	delete cover;

	if (context_changes == 0) {
		last_duration = (double)(fset->support) / 60 / 60;
	} else {
		last_duration = (double)(fset->support) / context_changes / 60 / 60;
	}
	return last_duration;
}

/**
 * Return the average druation of an association rule
 * @param arule The association rule.
 * @return The average duration of the association rule.
 */
double ContextAverageDurationConstraint::value(AssociationRule* arule) {
	Cover* cover = arule->itemset->getCover();

	vector<int> tids;
	for (int i = 0; i < cover->size; i++) {
		tids.push_back(cover->tid_table[i]);
	}
	sort(tids.begin(), tids.end());
	
	int context_changes = 0;
	for (int i = 0; i < tids.size(); i++) {
		if (i == 0) {
			context_changes++;
		} else if (tids[i] - 1 != tids[i-1]) {
			context_changes++;
		}
	}

	delete cover;

	if (context_changes == 0) {
		last_duration = (double)(arule->itemset->support) / 60 / 60;
	} else {
		last_duration = (double)(arule->itemset->support) / context_changes / 60 / 60;
	}
	return last_duration;
}

/**
 * Validate an itemset.
 * @param fset The itemset.
 * @return True if the average duration of the itemset is higher than the minimum
 * average duration of this constraint.
 */
bool ContextAverageDurationConstraint::validate(FITree* fset) {
	last_duration = value(fset);

	if (last_duration >= min_duration) {
		return true;
	}
	
	return false;
}

/**
 * Validate an association rule.
 * @param arule The association rule.
 * @return True if the average duration of the rule is higher than the minimum
 * average duration of this constraint.
 */
bool ContextAverageDurationConstraint::validate(AssociationRule* arule) {
	last_duration = value(arule);

	if (last_duration >= min_duration) {
		return true;
	}
	
	return false;
}

/**
 * Returns the minimum average duration.
 * @return The minimum average duration.
 */
double ContextAverageDurationConstraint::constraintLimit() const {
	return min_duration;
}

/**
 * Returns the name of this constraint.
 * @return The name.
 */
string ContextAverageDurationConstraint::name() const {
	return "Average Duration";
}

/**
 * Returns the abbreviation of this constraint.
 * @return The abbreviation.
 */
string ContextAverageDurationConstraint::abbreviation() const {
	return "avgd";
}

/**
 * Returns the description of this constraint.
 * @return The description.
 */
string ContextAverageDurationConstraint::description() const {
	return "Average duration is the average duration of a context in hours.";
}
