/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "ContextFrequencyConstraint.h"
#include "ContextTree.h"
#include "FITree.h"
#include "AssociationRule.h"

/**
 * Create a context frequency constraint with minimum frequency of 0.
 */
ContextFrequencyConstraint::ContextFrequencyConstraint() {
	last_frequency = 0;
	min_frequency = 0;
	null_duration = 0;
}

/**
 * Create a context frequency constraint with a minimum frequency.
 * @param d The minimum frequency.
 */
ContextFrequencyConstraint::ContextFrequencyConstraint(double d) {
	last_frequency = 0;
	min_frequency = d;
	null_duration = 0;
}

/**
 * Set the minimum frequency of this constraint.
 * @param d The minimum frequency.
 */
void ContextFrequencyConstraint::setMinFrequency(double d) {
	min_frequency = d;
}

/**
 * Set the context dataset.
 * @param ct The context data.
 */
void ContextFrequencyConstraint::setContextTree(ContextTree* ct) {
	context_tree = ct;
	null_duration = ct->nullDuration();
}

/**
 * Returns the result of last evaluation.
 * @return The frequency of the pattern evaluated last.
 */
double ContextFrequencyConstraint::lastResult() {
	return last_frequency;
}

/**
 * Return the context frequency of an itemset.
 * @param fset The itemset.
 * @return The frequency of the itemset.
 */
double ContextFrequencyConstraint::value(FITree* fset) {
	last_frequency = (double)fset->support / (double)(fset->root()->support - null_duration);
	return last_frequency;
}

/**
 * Return the context frequency of an association rule
 * @param arule The association rule.
 * @return The frequency of the association rule.
 */
double ContextFrequencyConstraint::value(AssociationRule* arule) {
	last_frequency = (double)arule->itemset->support / (double)(arule->itemset->root()->support - null_duration);
	return last_frequency;
}

/**
 * Validate an itemset.
 * @param fset The itemset.
 * @return True if the frequency of the itemset is higher than the minimum
 * frequency of this constraint.
 */
bool ContextFrequencyConstraint::validate(FITree* fset) {
	
	last_frequency = value(fset);

	if (last_frequency >= min_frequency) {
		return true;
	}
	
	return false;
}

/**
 * Validate an association rule.
 * @param arule The association rule.
 * @return True if the frequency of the rule is higher than the minimum
 * frequency of this constraint.
 */
bool ContextFrequencyConstraint::validate(AssociationRule* arule) {
	last_frequency = value(arule);

	if (last_frequency >= min_frequency) {
		return true;
	}
	
	return false;
}

/**
 * Returns the minimum frequency.
 * @return The minimum frequency.
 */
double ContextFrequencyConstraint::constraintLimit() const {
	return min_frequency;
}

/**
 * Returns the name of this constraint.
 * @return The name.
 */
string ContextFrequencyConstraint::name() const {
	return "Context Frequency";
}

/**
 * Returns the abbreviation of this constraint.
 * @return The abbreviation.
 */
string ContextFrequencyConstraint::abbreviation() const {
	return "f";
}

/**
 * Returns the description of this constraint.
 * @return The description.
 */
string ContextFrequencyConstraint::description() const {
	return "Relative context frequency is the duration of the pattern divided by the total duration from which we have information on.";
}
