/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "ContextMedianDurationConstraint.h"
#include "ContextTree.h"
#include "FITree.h"
#include "AssociationRule.h"
#include "Cover.h"

/**
 * Create a context frequency constraint with minimum frequency of 0.
 */
ContextMedianDurationConstraint::ContextMedianDurationConstraint() {
	last_duration = 0;
	min_duration = 0;
}

/**
 * Create a context frequency constraint with a minimum frequency.
 * @param d The minimum median duration.
 */
ContextMedianDurationConstraint::ContextMedianDurationConstraint(double d) {
	last_duration = 0;
	min_duration = d;
}

/**
 * Set the minimum median duration of this constraint.
 * @param d The minimum median duration.
 */
void ContextMedianDurationConstraint::setMinDuration(double d) {
	min_duration = d;
}

/**
 * Set the context dataset.
 * @param ct The context data.
 */
void ContextMedianDurationConstraint::setContextTree(ContextTree* ct) {
	context_tree = ct;
}

/**
 * Returns the result of last evaluation.
 * @return The median duration of the pattern evaluated last.
 */
double ContextMedianDurationConstraint::lastResult() {
	return last_duration;
}

/**
 * Return the median duration of an itemset.
 * @param fset The itemset.
 * @return The median duration of the itemset.
 */
double ContextMedianDurationConstraint::value(FITree* fset) {
	Cover* cover = fset->getCover();

	vector<int> tids;
	for (int i = 0; i < cover->size; i++) {
		tids.push_back(cover->tid_table[i]);
	}
	sort(tids.begin(), tids.end());
	
	// Root node with no items.
	if (tids.size() == 0) {
		last_duration = fset->support / 60 / 60;
		return last_duration;
	}
	
	vector<double> durations;
	for (int i = 0; i < tids.size(); i++) {
		int sum = context_tree->contextSupport(tids[i]);
		while ((i + 1) < tids.size() && tids[i] + 1 == tids[i+1]) {
			i++;
			sum = sum + context_tree->contextSupport(tids[i]);
		}
		durations.push_back(sum);
	}
	

	delete cover;

	sort(durations.begin(), durations.end());

	if (durations.size() % 2 == 0) {
		last_duration = ((double)durations[(int)(durations.size() / 2)] + (double)durations[(int)(durations.size() / 2) + 1]) / 2 / 60 / 60;
	} else {
		last_duration = (double)durations[(int)(durations.size() / 2)] / 60 / 60;
	}
	return last_duration;
}

/**
 * Return the median druation of an association rule
 * @param arule The association rule.
 * @return The median duration of the association rule.
 */
double ContextMedianDurationConstraint::value(AssociationRule* arule) {
	return value(arule->itemset);
}

/**
 * Validate an itemset.
 * @param fset The itemset.
 * @return True if the median duration of the itemset is higher than the minimum
 * median duration of this constraint.
 */
bool ContextMedianDurationConstraint::validate(FITree* fset) {
	last_duration = value(fset);

	if (last_duration >= min_duration) {
		return true;
	}
	
	return false;
}

/**
 * Validate an association rule.
 * @param arule The association rule.
 * @return True if the median duration of the rule is higher than the minimum
 * median duration of this constraint.
 */
bool ContextMedianDurationConstraint::validate(AssociationRule* arule) {
	last_duration = value(arule);

	if (last_duration >= min_duration) {
		return true;
	}
	
	return false;
}

/**
 * Returns the minimum median duration.
 * @return The minimum median duration.
 */
double ContextMedianDurationConstraint::constraintLimit() const {
	return min_duration;
}

/**
 * Returns the name of this constraint.
 * @return The name.
 */
string ContextMedianDurationConstraint::name() const {
	return "Median Duration";
}

/**
 * Returns the abbreviation of this constraint.
 * @return The abbreviation.
 */
string ContextMedianDurationConstraint::abbreviation() const {
	return "medd";
}

/**
 * Returns the description of this constraint.
 * @return The description.
 */
string ContextMedianDurationConstraint::description() const {
	return "Median duration is the median of durations of a context in hours.";
}
