/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <fstream>
using std::ifstream;

#include <algorithm>
using std::sort;

#include "ContextTree.h"
#include "Attribute.h"
#include "Item.h"
#include "Cover.h"
#include "FITree.h"

// Standard output function
ostream& operator<<(ostream& os, const ContextTree& tree) {
	os << "Duration attribute:" << endl;
	os << *(tree.attributes[tree.duration_attribute_index]) << endl;
	os << "Total duration:" << endl;
	os << tree.durations[tree.durations.size() - 1] << endl;
	os << "Attributes:" << endl;
	for(int i = 0; i < tree.attributes.size(); i++) {
		os << *(tree.attributes[i]) << endl;
	}

	return os;
}

/**
 * Create an empty context tree.
 */
ContextTree::ContextTree() {
	transactions = 0;
	duration_attribute_index = -1;
}

/**
 * Create a context tree from a text file.
 * @param data_file The name of the file.
 * @param duration_attribute The name of the time attribute.
 * @param column_separator The character separating the columns.
 * @param transaction_separator The characted separating the transactions.
 */
ContextTree::ContextTree(string data_file, string duration_attribute, char column_separator, char transaction_separator) {
	transactions = 0;
	duration_attribute_index = -1;
	buildTree(data_file, duration_attribute, column_separator, transaction_separator);
}

/**
 * The destructor.
 */
ContextTree::~ContextTree() {
	for (int i = 0; i < attributes.size(); i++) {
		delete attributes[i];
	}
}

/**
 * Build a context tree from a text file.
 * @param file_name The name of the file.
 * @param duration_attribute The name of the time attribute.
 * @param column_separator The character separating the columns.
 * @param transaction_separator The characted separating the transactions.
 */
void ContextTree::buildTree(string file_name, string duration_attribute, char column_separator, char transaction_separator) {
	buildTree(file_name.c_str(), duration_attribute, column_separator, transaction_separator);
}

/**
 * Build a context tree from a text file.
 * @param file_name The name of the file.
 * @param duration_attribute The name of the time attribute.
 * @param column_separator The character separating the columns.
 * @param transaction_separator The characted separating the transactions.
 */
void ContextTree::buildTree(const char* file_name, string duration_attribute, char column_separator, char transaction_separator) {
	ifstream database(file_name);
	
	if (!database.is_open()) {
		std::cerr << endl << "Error opening file: " << file_name << endl;
		exit(1);
	}

	char c;

	string value;
	int column = 0;
		
	int BUFFER_SIZE = 1000000;
	char* buffer = new char[BUFFER_SIZE];
	
	database.getline(buffer, BUFFER_SIZE, transaction_separator);
	int chars_read = database.gcount();
	
	// Read the attributes (headers of columns)
	for (int i = 0; i < chars_read - 1; i++) {
		c = buffer[i];
		//std::cerr << "c:" << (int)c << std::endl;
		
		if (c == column_separator) {
			if (value == duration_attribute) {
				duration_attribute_index = column;
			}
			attributes.push_back(new Attribute(value, column));
			column++;
			value = "";
		} else {
			value.append(1, c);
		}
	}
	
	if (value == duration_attribute) {
		duration_attribute_index = column;
	}
	attributes.push_back(new Attribute(value, column));

	if (duration_attribute_index == -1) {
		std::cerr << endl << "Error parsing file: " << file_name << ". Could not find duration attribute" << endl;
		exit(1);	
	}
	
	// Read the transactions
	value = "";
	column = 0;
	
	do {
		database.read(buffer, BUFFER_SIZE);
		chars_read = database.gcount();
		
		for (int i = 0; i < chars_read; i++) {
			c = buffer[i];
			
			if (c == column_separator) {
				//std::cerr << "Adding: v:'" << value << "' t:'" << transaction << "'" << endl;
				if (column == duration_attribute_index) {
					durations.push_back(atoi(value.c_str()));
				} else if (value != "") {
					attributes[column]->addValue(value, transactions);
				}
				column++;
				value = "";
			} else if (c == transaction_separator && column > 0) {
				//std::cerr << "Adding: v:'" << value << "' t:'" << transaction << "'" << endl;
				if (column == duration_attribute_index) {
					durations.push_back(atoi(value.c_str()));
				} else if (value != "") {
					attributes[column]->addValue(value, transactions);
				}
				column = 0;
				transactions++;
				value = "";
			} else {
				value.append(1, c);
			}
		}
	} while (!database.eof());
	
	database.close();
}


/**
 * Count the duration of a context given the cover of the context.
 * @param cover The cover.
 * @return The summed duration as given by the time-column
 */
int ContextTree::contextSupport(Cover* cover) {
	int support = 0;
	int tid;
	
	for (int i = 0; i < cover->size; i++) {
		tid = cover->tid_table[i];
		if (tid == 0) {
			support = support + durations[tid];
		} else {
			support = support + (durations[tid] - durations[tid - 1]);
		}
	}
	
	return support;
}

/**
 * Count the duration of a context given the cover of the context.
 * @param tid The transaction identifier
 * @return The duration of tid given by the time-column
 */
int ContextTree::contextSupport(int tid) {
	if (tid == 0) {
		return durations[tid];
	} else {
		return (durations[tid] - durations[tid - 1]);
	}
}

/**
 * Return the combined duration during which there has been no available context variables.
 * @return The duration during which there has been no available context variables.
 */
int ContextTree::nullDuration() {

	Cover* c = new Cover();
	for (int i = 0; i < transactions; i++) {
		c->addTransaction(i);
	}
	
	Cover* prev_c = c;
	
	for (int i = 0; i < attributes.size(); i++) {
		for (map<string, Item*>::iterator iter = attributes[i]->items.begin(); iter != attributes[i]->items.end(); iter++) {
			prev_c = c;
			c = differenceCover(c, iter->second->cover);
			delete prev_c;
		}
	}

	int support = contextSupport(c);
	delete c;
		
	return support;
}

/**
 * Return the maximum duration of this context dataset
 */
int ContextTree::getTransactions() {
	return durations[durations.size() - 1];
}


/**
 * Get the frequent items with certain minimum support.
 * @param min_supp The minimum absolute support.
 * @return The FITree containing all frequent items.
 */
FITree* ContextTree::frequentItems(int min_supp) {
	FITree* root = new FITree();
	root->support = durations[durations.size() - 1] - nullDuration();
	root->height = 0;
	root->upper_bound = durations[durations.size() - 1] - nullDuration();
	root->lower_bound = 0;
	
	FITree* item;
	int support;
	
	for (int i = 0; i < attributes.size(); i++) {
		for (map<string, Item*>::iterator iter = attributes[i]->items.begin(); iter != attributes[i]->items.end(); iter++) {
			support = contextSupport(iter->second->cover);
			
			if (support >= min_supp) {
				item = new FITree(iter->second, root, iter->second->cover, support);
				item->height = 1;
				item->upper_bound = root->upper_bound;
				item->lower_bound = 0;
				item->tidhash = iter->second->cover->tidhash();
				root->addChild(item);
			}
		}
	}
	
	// Sorting the initial 1-itemsets in ascending order of support will result in
	// more cases of support of a superset being equal to the support of the itemset
	
	// using the upper will result in a bug in ndi-algorithm, since the findItem binary search relies on *item (cover->size) compare, not support compare...
	//sort((root->children).begin(), (root->children).end(), fi_tree_pointer_support_compare);

	sort((root->children).begin(), (root->children).end(), fi_tree_pointer_compare);

	return root;
}
