/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef FIRM_CONTEXTTREE
#define FIRM_CONTEXTTREE

#include <iostream>
using std::ostream;
using std::cout;
using std::endl;

#include <vector>
using std::vector;

#include <string>
using std::string;

class Cover;
class Item;
class Attribute;
class FITree;

/**
 * A context tree is a regular attribute-value dataset with
 * each transaction having a special time attribute telling
 * the duration of the context that is described by the other
 * variables. A context tree can be formed from for example data given below:
 *
 * TIME    BASE    IDLE    VOLUME  ACTIVE_APPLICATION
 * 10      2       idle    2       none
 * 140     1       idle    2       none
 * 3000
 * 3400    2       active  7       phonebook
 * 3800    2       active  7       phonebook
 * 3830    3       idle    2       none
 *
 * Now the TIME column describes how long the context has been valid.
 * Given in time-units from the start of the measurement. For example,
 * the second line indicates that from time 11 to 140 user has been in
 * context: base=2, idle= idle, volume=2, active_application=none.
 *
 * @author Juho Muhonen
 * @modification 7.11.2005 contextSupport(int) for a single tid
 * @modification 26.10.2005 root node now has the time during which we have at
 *                          least some knowledge as it's support that is max_duration - nullDuration()
 * @modification 10.10.2005 commented all
 */
class ContextTree {
private:
	/**
	 * The attributes (the column names)
	 */
	vector<Attribute*> attributes;
	
	/**
	 * The number of transactions (lines)
	 */
	int transactions;
	
	/**
	 * The times in the special time column.
	 */
	vector<int> durations;
	
	/**
	 * Index of the time column.
	 */
	int duration_attribute_index;
	
public:
	/**
	 * Create an empty context tree.
	 */
	ContextTree();
	
	/**
	 * Create a context tree from a text file.
	 */
	ContextTree(string, string, char, char);
	
	/**
	 * The destructor.
	 */
	~ContextTree();

	/**
	 * Build a context tree from a text file.
	 */
	void buildTree(string, string, char, char);

	/**
	 * Build a context tree from a text file.
	 */
	void buildTree(const char*, string, char, char);
	
	/**
	 * Count the duration of a context given the cover of the context.
	 */
	int contextSupport(Cover*);
	
	/**
	 * Return the duration of a context given the context tid.
	 */
	int contextSupport(int);

	/**
	 * Return the combined duration during which there has
	 * been no available context variables.
	 */
	int nullDuration();

	/**
	 * Get the frequent items.
	 */
	FITree* frequentItems(int);
	
	/**
	 * Return the maximum duration of this context dataset
	 */
	int getTransactions();

	/**
	 * Output this tree to a stream.
	 */
	friend ostream& operator<<(ostream&, const ContextTree&);
	
};

ostream& operator<<(ostream&, const ContextTree&);
#endif
