/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "Cover.h"

/**
 * Outputs the cover to a stream.
 * @param os The output stream.
 * @param cov The cover.
 * @return The output stream.
 */
ostream& operator<<(ostream& os, const Cover& cov) {
	os << "tids: " << cov.size << " [";
	for(int i = 0; i < cov.size; i++) {
		os << cov.tid_table[i] << " ";
	}
	os << "]";
	return os;
}

int Cover::CREATIONS = 0;
int Cover::DELETIONS = 0;

/**
 * Creates a new cover with capacity of DEFAULT_CAPACITY
 */
Cover::Cover() {
	size = 0;
	capacity = DEFAULT_CAPACITY;
	tid_table = new int[DEFAULT_CAPACITY];
}

/**
 * Creates a new cover with given capacity.
 * @param c The capacity.
 */
Cover::Cover(int c) {
	Cover::CREATIONS++;
	
	size = 0;
	capacity = c;
	tid_table = new int[c];
}

/**
 * Copies a cover.
 * @param c The cover to be copied.
 */
Cover::Cover(const Cover& c) {
	Cover::CREATIONS++;

	size = c.size;
	capacity = c.capacity;
	tid_table = new int[capacity];
	for (int i = 0; i < size; i++) {
		tid_table[i] = c.tid_table[i];
	}
}

/**
 * The destructor.
 */
Cover::~Cover() {
	Cover::DELETIONS++;

	delete tid_table;
}

/**
 * Add a tid to this cover.
 * @param tid The transaction identifier.
 */
void Cover::addTransaction(int tid) {	
	if (size == capacity - 1) {
		capacity = capacity * 2;
		int* new_tid_table = new int[capacity];
		
		for (int i = 0; i < size; i++) {
			new_tid_table[i] = tid_table[i];
		}
		
		new_tid_table[size] = tid;
		size++;
		
		delete tid_table;
		tid_table = new_tid_table;
	} else {
		tid_table[size] = tid;
		size++;
	}
}

/**
 * Computes a tidhash of the cover. Tidhash is defined as the sum of all transaction identifiers.
 * @return The tidhash value. -1 if the cover is empty.
 */
int Cover::tidhash() {	
	if (size == 0) {
		return -1;
	}
	
	int tidhash = 0;
	for (int i = 0; i < size; i++) {
		tidhash = tidhash + tid_table[i];
	}
	
	return tidhash;
}

/**
 * Creates a new cover by unioning two covers. Note that the cover created this
 * way must be deleted eventually to avoid memory leakage.
 * @param c1 The first cover.
 * @param c2 The second cover.
 * @return The union cover.
 */
Cover* unionCover(const Cover* c1, const Cover* c2){
	Cover* union_cover = new Cover(c1->size + c2->size);
	
	int c1_iter = 0;
	int c2_iter = 0;
	
	while (c1_iter < c1->size && c2_iter < c2->size) {
		if (c1->tid_table[c1_iter] < c2->tid_table[c2_iter]) {
			union_cover->addTransaction(c1->tid_table[c1_iter]);
			c1_iter++;
		} else if (c1->tid_table[c1_iter] > c2->tid_table[c2_iter]) {
			union_cover->addTransaction(c2->tid_table[c2_iter]);
			c2_iter++;
		} else {
			union_cover->addTransaction(c1->tid_table[c1_iter]);
			c1_iter++;
			c2_iter++;
		}
	}
		
	while (c1_iter < c1->size) {
		union_cover->addTransaction(c1->tid_table[c1_iter]);
		c1_iter++;			
	}

	while (c2_iter < c2->size) {
		union_cover->addTransaction(c2->tid_table[c2_iter]);
		c2_iter++;
	}

	return union_cover;
}

//Cover* unionCover(Cover** cc, int size) {
//	return new Cover();
//}

/**
 * Creates a new cover by intersecting two covers. Note that the cover created this
 * way must be deleted eventually to avoid memory leakage.
 * @param c1 The first cover.
 * @param c2 The second cover.
 * @return The intersection of the two covers.
 */
Cover* intersectionCover(const Cover* c1, const Cover* c2) {
	Cover* intersection_cover = new Cover(c1->size);
	//Cover* intersection_cover = Cover::_store_cover;
	//intersection_cover->size = 0;

	int c1_iter = 0;
	int c2_iter = 0;
	
	while (c1_iter < c1->size && c2_iter < c2->size) {
		if (c1->tid_table[c1_iter] < c2->tid_table[c2_iter]) {
			c1_iter++;
		} else if (c1->tid_table[c1_iter] > c2->tid_table[c2_iter]) {
			c2_iter++;
		} else {
			intersection_cover->addTransaction(c1->tid_table[c1_iter]);
			c1_iter++;
			c2_iter++;
		}
	}

	return intersection_cover;
}

/**
 * Creates a new cover by intersecting multiple covers. Note that the cover created this
 * way must be deleted eventually to avoid memory leakage.
 * @param cc The array of cover pointers.
 * @param size The size of the array.
 * @return The intersection of covers.
 */
Cover* intersectionCover(Cover** cc, int size) {
	if (size == 0) {
		return new Cover();
	} else if (size == 1) {
		return new Cover(*(cc[0]));
	}
	
	Cover* intersection_cover = intersectionCover(cc[0], cc[1]);
	Cover* temp_cover = intersection_cover;
	
	for (int i = 2; i < size; i++) {
		intersection_cover = intersectionCover(intersection_cover, cc[i]);
		delete temp_cover;
		temp_cover = intersection_cover;
	}
	
	return intersection_cover;
	
	/*Cover* intersection_cover = new Cover(cc[0]->size);
	
	bool iterators_valid = true;
	
	int* iterator = new int[size];
	for (int i = 0; i < size; i++) {
		iterator[i] = 0;
		if (cc[i]->size == 0) {
			iterators_valid = false;
		}
	}
	
	int smallest_iter;
	int smallest_tid;
	int same;
	while (iterators_valid) {
		// Find the smallest tid
		smallest_tid = cc[0]->tid_table[iterator[0]];
		same = 1;
		for (int i = 1; i < size; i++) {
			if (cc[i]->tid_table[iterator[i]] < smallest_tid) {
				smallest_tid = cc[i]->tid_table[iterator[i]];
				smallest_iter = i;
			} else if (cc[i]->tid_table[iterator[i]] == smallest_tid) {
				same++;
			}
		}
		
		// If all are smallest, add the transaction and increase all iterators
		// otherwise increase the iterator with smallest tid
		if (same == size) {
			intersection_cover->addTransaction(smallest_tid);
			for (int i = 0; i < size; i++) {
				iterator[i]++;
				if (cc[i]->size == iterator[i]) {
					iterators_valid = false;
				}
			}
		} else {
			iterator[smallest_iter]++;
			if (cc[smallest_iter]->size == iterator[smallest_iter]) {
				iterators_valid = false;
			}
		}
	}
	
	delete iterator;
	return intersection_cover;*/
}

/**
 * Creates a new cover by taking the difference of two covers. Note that the cover created this
 * way must be deleted eventually to avoid memory leakage.
 * @param c1 The first cover.
 * @param c2 The second cover.
 * @return The difference (c1 - c2) of the covers.
 */
Cover* differenceCover(const Cover* c1, const Cover* c2){
	Cover* diff_cover = new Cover(c1->size);

	int c1_iter = 0;
	int c2_iter = 0;
		
	while (c1_iter < c1->size && c2_iter < c2->size) {
		if (c1->tid_table[c1_iter] < c2->tid_table[c2_iter]) {
			diff_cover->addTransaction(c1->tid_table[c1_iter]);
			c1_iter++;
		} else if (c1->tid_table[c1_iter] > c2->tid_table[c2_iter]) {
			c2_iter++;
		} else {
			c1_iter++;
			c2_iter++;
		}
	}
	
	while (c1_iter < c1->size) {
		diff_cover->addTransaction(c1->tid_table[c1_iter]);
		c1_iter++;			
	}
	
	return diff_cover;
}

/*int main() {
	Cover* cov1 = new Cover();
	cov1->addTransaction(0);
	cov1->addTransaction(1);
	cov1->addTransaction(2);
	cov1->addTransaction(3);
	cov1->addTransaction(4);

	Cover* cov2 = new Cover();
	cov2->addTransaction(3);
	cov2->addTransaction(4);
	cov2->addTransaction(5);
	cov2->addTransaction(7);

	Cover* cov3 = new Cover();
	cov3->addTransaction(7);
	cov3->addTransaction(8);
	cov3->addTransaction(9);
	cov3->addTransaction(10);
	cov3->addTransaction(100);
	cov3->addTransaction(10234);

	std::cerr << "Cov1: " << *cov1 << std::endl;
	std::cerr << "Cov2: " << *cov2 << std::endl;
	std::cerr << "Cov3: " << *cov3 << std::endl;
	
	std::cerr << "Cov1 - Cov2: " << *(differenceCover(cov1, cov2)) << std::endl;
	std::cerr << "Cov1 - Cov3: " << *(differenceCover(cov1, cov3)) << std::endl;
	std::cerr << "Cov2 - Cov1: " << *(differenceCover(cov2, cov1)) << std::endl;
	
	std::cerr << "Cov1 union cov2: " << *(unionCover(cov1, cov2)) << std::endl;
	std::cerr << "Cov1 union cov3: " << *(unionCover(cov1, cov3)) << std::endl;

	std::cerr << "Cov2 union cov3: " << *(unionCover(cov2, cov3)) << std::endl;
	std::cerr << "Cov3 union cov2: " << *(unionCover(cov3, cov2)) << std::endl;

	std::cerr << "Cov1 intersection cov2: " << *(intersectionCover(cov1, cov2)) << std::endl;
	std::cerr << "Cov1 intersection cov3: " << *(intersectionCover(cov1, cov3)) << std::endl;
	std::cerr << "Cov2 intersection cov3: " << *(intersectionCover(cov2, cov3)) << std::endl;
	std::cerr << "Cov3 intersection cov2: " << *(intersectionCover(cov3, cov2)) << std::endl;
	
	Cover** covers;
	covers = new Cover*[3];
	covers[0] = cov1;
	covers[1] = cov2;
	covers[2] = cov3;
	
	std::cerr << "cov1 intersection Cov3 intersection cov2: " << *(intersectionCover(covers, 3)) << std::endl;
	//Cover* cover_table2[2];
	//cover_table2[0] = cov1;
	//cover_table2[1] = cov2;
	//cover_table2[2] = cov3;

	//std::cerr << "Cov1 union cov2 : " << *(unionCover(cover_table2, 2)) << std::endl;
	//std::cerr << "Cov1 intersection cov2 : " << *(intersectionCover(cover_table2, 2)) << std::endl;

	//Cover* cover_table[3];
	//cover_table[0] = cov1;
	//cover_table[1] = cov2;
	//cover_table[2] = cov3;

	//std::cerr << "Cov1 union cov2 union cov3: " << *(unionCover(cover_table, 3)) << std::endl;
	//std::cerr << "Cov1 intersection cov2 intersection cov3: " << *(intersectionCover(cover_table, 3)) << std::endl;
}*/
