/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef FIRM_COVER
#define FIRM_COVER

#define DEFAULT_CAPACITY 64

#include <iostream>
using std::ostream;

/**
 * A class for storing and manipulating an itemset cover. A cover is a set of transaction identifiers (tids).
 * This class provides fast funtionality for computing unions, intersections and differences of covers.
 * For efficiency this class _will not_ perform any ordering of the tids, but rather it assumes that
 * the tids are given in increasing order.
 * @author Juho Muhonen
 * @modification 10.10.2005 commented all
 * @modification 20.9.2005 tidhash()
 * @modification 23.3.2004 created
 */
class Cover {
public:
	/**
	 * The number of tids in the cover.
	 */
	int size;
	
	/**
	 * The number of tids the cover can hold before it has to allocate more memory.
	 */
	int capacity;
	
	/**
	 * The array holding the tids.
	 */
	int* tid_table;
	
public:
	/**
	 * A static variable for detecting memory leaks (debugging purposes).
	 */
	static int CREATIONS;

	/**
	 * A static variable for detecting memory leaks (debugging purposes).
	 */
	static int DELETIONS;
	
	/**
	 * Create an empty cover.
	 */
	Cover();

	/**
	 * Create an empty cover with given capacity.
	 */
	Cover(int);
	
	/**
	 * Copies a cover.
	 */
	Cover(const Cover&);

	/**
	 * The destructor.
	 */
	~Cover();

	/**
	 * Adds a transaction identifier to this cover.
	 */
	void addTransaction(int);

	/**
	 * Computes a tidhash of the cover. Tidhash is defined as the sum of all transaction identifiers.
	 */
	int tidhash();
	
	/**
	 * Creates a new cover by unioning two covers.
	 */
	friend Cover* unionCover(const Cover*, const Cover*);

	/**
	 * Creates a new cover by unioning given number of covers
	friend Cover* unionCover(Cover**, int);
	 */

	/**
	 * Creates a new cover by intersecting two covers.
	 */
	friend Cover* intersectionCover(const Cover*, const Cover*);

	/**
	 * Creates a new cover by intersecting given number of covers.
	 */
	friend Cover* intersectionCover(Cover**, int);

	/**
	 * Creates a new cover by taking the difference of two covers.
	 */
	friend Cover* differenceCover(const Cover*, const Cover*);

	/**
	 * Outputs the cover.
	 */
	friend ostream& operator<<(ostream& os, const Cover& cov);
};

Cover* unionCover(const Cover*, const Cover*);
Cover* unionCover(Cover**, int);

Cover* intersectionCover(const Cover*, const Cover*);
Cover* intersectionCover(Cover**, int);

Cover* differenceCover(const Cover*, const Cover*);

ostream& operator<<(ostream& os, const Cover& cov);
#endif
