/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <queue>
using std::queue;

#include <algorithm>
using std::sort;

#include "AssociationRule.h"
#include "FITree.h"
#include "Attribute.h"
#include "Item.h"
#include "Cover.h"

/**
 * Outputs a FITree to an output stream
 * @param os The output stream.
 * @param node The tree.
 * @return The output stream.
 */
ostream& operator<<(ostream& os, const FITree& node) {
	const FITree* iter = &node;
	while (iter != NULL) {
		for (int i = 0; i < iter->items.size(); i++) {
			os << *(iter->items[i]) << " ";
		}
		iter = iter->parent;
	}
	os << "";
}

int FITree::CREATIONS = 0;
int FITree::DELETIONS = 0;

/**
 * Compares two FITrees (nodes). By the first item stored in them.
 * @param is2 The node this node is compared to.
 * @return zero if is2 is smaller than this. non-zero otherwise.
 */
int FITree::operator<(const FITree& is2) const {
	return (*(this->items[0]) < *(is2.items[0]));
}

/**
 * Creates a new root node.
 */
FITree::FITree() {
	FITree::CREATIONS++;
	
	parent = NULL;
	cover = NULL;
	support = 0;
	height = 0;
	tidhash = 0;
	upper_bound = INT_MAX;
	lower_bound = 0;
#ifdef _STORE_SUBSETS
	_subsets = NULL;
#endif
}

/**
 * Creates a copy of the tree.
 * @param original The tree to be copied.
 */
FITree::FITree(const FITree& original) {
	FITree::CREATIONS++;
	
	parent = original.parent;
	cover = original.cover;
	support = original.support;
	items = original.items;
	height = original.height;
	tidhash = original.tidhash;
	upper_bound = original.upper_bound;
	lower_bound = original.lower_bound;
	
	for (int i = 0; i < original.children.size(); i++) {
		addChild(new FITree(*(original.children[i])));
	}
}

/**
 * Creates a new root node with an item stored.
 * @param i The item to be stored.
 */
FITree::FITree(Item* i) {
	FITree::CREATIONS++;
	
	items.push_back(i);
	parent = NULL;
	cover = NULL;
	support = 0;
	height = 0;
	tidhash = 0;
	upper_bound = INT_MAX;
	lower_bound = 0;
#ifdef _STORE_SUBSETS
	_subsets = NULL;
#endif
}

/**
 * Creates a new node with parent and an item stored.
 * @param i The item
 * @param p The parent
 */
FITree::FITree(Item* i, FITree* p) {
	FITree::CREATIONS++;
	
	items.push_back(i);
	parent = p;
	cover = NULL;
	support = 0;
	height = p->height + 1;
	tidhash = 0;
	upper_bound = INT_MAX;
	lower_bound = 0;
#ifdef _STORE_SUBSETS
	_subsets = NULL;
#endif
}

/**
 * Creates a new node with parent and cover and an item stored.
 * @param i The item
 * @param p The parent
 * @param c The cover
 */
FITree::FITree(Item* i, FITree* p, Cover* c) {
	FITree::CREATIONS++;
	
	items.push_back(i);
	parent = p;
	cover = c;
	support = 0;
	height = p->height + 1;
	tidhash = 0;
	upper_bound = INT_MAX;
	lower_bound = 0;
#ifdef _STORE_SUBSETS
	_subsets = NULL;
#endif
}

/**
 * Creates a new node with parent, cover and support and an item stored.
 * @param i The item
 * @param p The parent
 * @param c The cover
 * @param s The support
 */
FITree::FITree(Item* i, FITree* p, Cover* c, int s) {
	FITree::CREATIONS++;
	
	items.push_back(i);
	parent = p;
	cover = c;
	support = s;
	height = p->height + 1;
	tidhash = 0;
	upper_bound = INT_MAX;
	lower_bound = 0;
#ifdef _STORE_SUBSETS
	_subsets = NULL;
#endif
}

/**
 * The destructor.
 */
FITree::~FITree() {
	FITree::DELETIONS++;
		
	for (int i = 0; i < children.size(); i++) {
		delete children[i];
	}

	// Since when stored an association rule is stored both in cond and cons
	// rules it is sufficient (and correct) to delete only the other
	map<FITree*, AssociationRule*>::iterator iter;
	for(iter = condition_rules.begin(); iter != condition_rules.end(); iter++) {
		delete iter->second;
	}
#ifdef _STORE_SUBSETS
	if (_subsets != NULL) {
		delete _subsets;
	}
#endif
}


/**
 * Tries to find an item from this node.
 * @param child The searched item.
 * @return This is returned if the node contains the item, NULL otherwise.
 */
FITree* FITree::findItem(Item* child) {
	// Check the nodes itemlist for item
	if (items.size() <= 3) {
		for (int i = 0; i < items.size(); i++) {
			if (items[i] == child) {
				return this;
			}
		}
	} else {
		// Do a binary search on items.
		int upper_bound = items.size();
		int lower_bound = 0;
		int middle = (upper_bound + lower_bound) / 2;
		while (upper_bound - lower_bound > 5) {
		
			if (child == items[middle]) {
				return this;
			} else if (*child < *(items[middle])) {
				upper_bound = middle;
				middle = (upper_bound + lower_bound) / 2;
			} else {
				lower_bound = middle;
				middle = (upper_bound + lower_bound) / 2;
			}
		}
		for (int i = lower_bound; i < upper_bound; i++) {
			if (child == items[i]) {
				return this;
			}		
		}
	}
	
	return NULL;
}

/**
 * Tries to find the child containing a given item (or returns this node, if this node contains the item)
 * This function finds and returns a child FITree
 * With item child. NULL is returned if no such child is found
 * This function might return a pointer to the node it was called from
 * if the node contains the given item.
 * NOTE: does not work with closed itemset tree.
 * @param child The serached item
 * @return This if this node contains the item, a child containing the item or NULL if neither this or a child contains the item.
 */
FITree* FITree::findChild(Item* child) {
	FITree* temp = findItem(child);
	if (temp != NULL) {
		return temp;
	}
	
	// If item is not in itemlist check the actual children.
	if (children.size() <= 3) {
		for (int i = 0; i < children.size(); i++) {
			if (child == (children[i])->items[0]) {
				return children[i];
			}
		}
	} else {
		// Do a binary search on children.
		int upper_bound = children.size();
		int lower_bound = 0;
		int middle = (upper_bound + lower_bound) / 2;
		while (upper_bound - lower_bound > 5) {
		
			if (child == (children[middle])->items[0]) {
				return children[middle];
			} else if (*child < *((children[middle])->items[0])) {
				upper_bound = middle;
				middle = (upper_bound + lower_bound) / 2;
			} else {
				lower_bound = middle;
				middle = (upper_bound + lower_bound) / 2;
			}
		}
		for (int i = lower_bound; i < upper_bound; i++) {
			if (child == (children[i])->items[0]) {
				return children[i];
			}		
		}
	}
	return NULL;
}

/**
 * Tries to find an itemset using this node as the root.
 * @param itemset The items in the searched itemset (have to be ordered)
 * @return The itemset, NULL if the itemset could not be found.
 */
FITree* FITree::findItemset(vector<Item*>* itemset) {
	FITree* finode = this;
	for (int i = 0; i < itemset->size(); i++) {
		if (finode == NULL) {
			return NULL;
		}
		finode = finode->findChild((*itemset)[i]);
	}
	return finode;
}


/** 
 * Returns the root of the tree this node belongs to.
 * @return The root of the tree.
 */
FITree* FITree::root() {
	FITree* finode = this;
	while (finode->parent != NULL) {
		finode = finode->parent;
	}
	return finode;
}

/**
 * Finds an union itemset of two itemsets from the tree.
 * @param itemset The itemset this itemset is unioned with.
 * @return The union itemset.
 */
FITree* FITree::unionItemset(FITree* itemset) {
	vector<Item*>* union_items = new vector<Item*>();
	
	FITree* finode = this;
	while (finode->parent != NULL) {
		for (int i = 0; i < items.size(); i++) {
			union_items->push_back(items[i]);
		}
		finode = finode->parent;
	}

	finode = itemset;
	while (finode->parent != NULL) {
		for (int i = 0; i < items.size(); i++) {
			union_items->push_back(items[i]);
		}
		finode = finode->parent;
	}

	sort(union_items->begin(), union_items->end(), item_pointer_compare);
	finode = finode->findItemset(union_items);
	delete union_items;
	
	return finode;
}

/**
 * Adds a child to this node.
 * @param c The child node
 */
void FITree::addChild(FITree* c) {
	c->parent = this;
	children.push_back(c);
}

/**
 * Remove and return the latest child.
 * @return The latest child.
 */
FITree* FITree::removeChild() {
	FITree* child = children.back();
	children.pop_back();
	return child;
}

/**
 * Remove and return a child at location i
 * @return The i:th child
 */
FITree* FITree::removeChild(int i) {
	vector<FITree*>::iterator iter = children.begin();
	iter = iter + i;
	
	FITree* child = children[i];
	children[i]->parent = NULL;
	children.erase(iter);
	return child;
}

/**
 * Adds an item to this node.
 * @param i The item
 */
void FITree::addItem(Item* i) {
	items.push_back(i);
}

/**
 * Clears the items of this node.
 */
void FITree::clearItems() {
	items.clear();
}

/**
 * Clears the children of this node.
 */
void FITree::clearChildren() {
	children.clear();
}

/**
 * Checks if this itemset has an item with a given attribute.
 * @param attr The attribute
 * @return True if there is an item with the given attribute, false otherwise.
 */
bool FITree::hasAttribute(Attribute* attr) const {
	for (int i = 0; i < items.size(); i++) {
		if (items[i]->attribute == attr) {
			return true;
		}
	}
	
	if (parent == NULL) {
		return false;
	} else {
		return parent->hasAttribute(attr);
	}
}

/**
 * Checks if all the subsets of this itemset are frequent.
 * @return True if all the subsets of the itemset are frequent, false otherwise.
 */
bool FITree::subsetsFrequent() {
	vector<Item*> item_stack;
	FITree* node = this;
	FITree* iterator;
	
	// Two-sets are always frequent :) comment this out, if it isn't the case
	for (int i = 0; i < 3; i++) {
		if (node == NULL) {
			return true;
		}

		if ((node->items).size() != 0) {
			item_stack.push_back(node->items[0]);
		}

		node = node->parent;
	}
	
	while (node != NULL) {
		iterator = node;
		
		for (int i = item_stack.size() - 2; i >= 0; i--) {
			iterator = iterator->findChild(item_stack[i]);
			if (iterator == NULL) {
				return false;
			}
		}
		
		if ((node->items).size() != 0) {
			item_stack.push_back(node->items[0]);
		}
		
		node = node->parent;
	}
	
	return true;
}

/**
 * Actually computes and stores the support of this itemset using the covers of the stored items.
 */
void FITree::countSupport() {
	Cover** covers = new Cover*[height];

	FITree* node = this;
	int i = 0;
	while (node->parent != NULL) {
		covers[i] = node->items[0]->cover;
		i++;
		node = node->parent;
	}

	Cover* intersection = intersectionCover(covers, height);
	support = intersection->size;
	
	delete intersection;
	delete covers;
}

/**
 * Returns the cover of this itemset. That is, the intersection of the item covers.
 * @return The itemset cover
 */
Cover* FITree::getCover() {
	Cover** covers = new Cover*[height];

	FITree* node = this;
	int i = 0;
	while (node->parent != NULL) {
		covers[i] = node->items[0]->cover;
		i++;
		node = node->parent;
	}

	Cover* intersection = intersectionCover(covers, height);
	
	delete covers;
	return intersection;
}

/**
 * Finds and returns all association rules that can be produced from
 * this itemset. Note that the tree must be expanded before this function works.
 * @return A vector of all association rules that can be produced from this itemset.
 */
vector<AssociationRule*>* FITree::associationRules() {
	vector<Item*> itemset;
	FITree* node = this;
	while (node->parent != NULL) {
		itemset.push_back(node->items[0]);
		node = node->parent;
	}
	
	queue<AssociationRule*> q;
	q.push(new AssociationRule(this, node, node));
	AssociationRule* iterator;
	AssociationRule* arule_without_item;
	int queue_size = 1;
	
	// itemset.size() - 1 levels (items) in this search
	for (int i = itemset.size() - 1; i >= 0; i--) {
		// each level has 2^(itemset.size() - 1 - i) subsets
		for (int j = 0; j < queue_size; j++) {
			iterator = q.front();
			q.pop();

			arule_without_item = new AssociationRule(*iterator);

			iterator->condition = iterator->condition->findChild(itemset[i]);
			arule_without_item->consequent = arule_without_item->consequent->findChild(itemset[i]);
			
			q.push(iterator);
			q.push(arule_without_item);
		}
		queue_size = queue_size * 2;
	}	
	
	vector<AssociationRule*>* arules = new vector<AssociationRule*>();
	while(!q.empty()) {
		iterator = q.front();
		q.pop();
		arules->push_back(iterator);
	}
	
	return arules;
}

/**
 * Returns the number of items in this FITree node
 * @return The number of items in this node.
 */
int FITree::itemCount() {
	return items.size();
}

/**
 * Returns the number of items in this itemset.
 * @return The number of items.
 */ 
int FITree::size() {
	FITree* finode = this;
	int size = 0;
	
	while (finode != NULL) {
		size = size + finode->itemCount();
		finode = finode->parent;
	}
	return size;
}

/**
 * Tells whether this itemset is a superset of a given itemset.
 * @param subset The possible subset.
 * @return True if this itemset is a superset of the given subset, false otherwise.
 */
bool FITree::isSuperset(FITree* subset) {
	if (this->size() < subset->size()) {
		return false;
	}
	
	FITree* finode = subset;
	
	// For each node in subset
	while (finode != NULL) {
		// For each item in node
		for (int i = 0; i < finode->items.size(); i++) {
			// If the item is not found, this is not a superset of the given itemset
			if (!hasItem(finode->items[i])) {
				return false;
			}
		}
		finode = finode->parent;
	}
	
	return true;
}

/**
 * Checks if this itemset has an item.
 * @param item The item to be searched
 * @return True if this itemset has the item. False otherwise.
 */
bool FITree::hasItem(Item* item) {
	FITree* finode = this;
	
	while (finode != NULL) {
		if (finode->findItem(item) != NULL) {
			return true;
		}
		finode = finode->parent;
	}
	
	return false;
}

/**
 * Finds and returns all subsets of a node. Note that the tree must be
 * expanded before this function works.
 * @return A vector of subsets of this itemset.
 */
vector<FITree*>* FITree::subsets() {
#ifdef _STORE_SUBSETS
	if (_subsets != NULL) {
		return _subsets;
	}
#endif
	vector<FITree*>* subsets = new vector<FITree*>();

	vector<Item*>* itemset = new vector<Item*>();
	FITree* node = this;

	while (node->parent != NULL) {
		itemset->push_back(node->items[0]);
		node = node->parent;
	}
	node->subsets(itemset->size(), itemset, subsets);
	
	delete itemset;
#ifdef _STORE_SUBSETS
	_subsets = subsets;
#endif
	return subsets;
}

/**
 * A helper function for subsets()
 * @param i The iterator
 * @param itemset The vector to store the current itemset
 * @param subsets The vector to store the subsets
 */
void FITree::subsets(int i, vector<Item*>* itemset, vector<FITree*>* subsets) {
	if (i == 0) {
		subsets->push_back(this);
	} else {
		Item* item = (*itemset)[i - 1];
		FITree* child = this->findChild(item);
		if (child != NULL) {
			child->subsets(i - 1, itemset, subsets);
		} else {
		}
		this->subsets(i - 1,itemset, subsets);
	}
}

/**
 * Outputs all items from this node to the root of this tree.
 * @param os The output stream.
 */
void FITree::outputItems(ostream& os) {
	outputItems(os, NULL);
}

/**
 * Outputs all items from this node to a given node.
 * @param os The output stream.
 * @param root The "false root" if root is NULL all items are output
 */
void FITree::outputItems(ostream& os, FITree* root) {
	FITree* iterator = this;
	while (iterator != root) {
		for (int i = 0; i < (iterator->items).size(); i++) {
			os << *(iterator->items[i]) << " ";
		}
		iterator = iterator->parent;
	}
}

/**
 * Returns the number of frequent sets from this node onward
 * @return The number of itemsets from this node onward.
 */
int FITree::numberOfFrequentSets() const {
	int fsets = 0;
	int multiplier = 1;
	
	int exponent;
	if (items.size() == 0) {
	  exponent = 0;
	} else {
	  exponent = items.size() - 1;	
	}
	
	for (int i = 0; i < exponent; i++) {
		multiplier = multiplier * 2;
	}
	
	fsets = multiplier;
	
	for (int i = 0; i < children.size(); i++) {
		fsets = fsets + multiplier * children[i]->numberOfFrequentSets();
	}
	
	return fsets;
}

/**
 * Returns the number of nodes from this node down.
 * @return The number of nodes
 */
int FITree::numberOfNodes() const {
	int nodes = 1;
	
	for (int i = 0; i < children.size(); i++) {
		nodes = nodes + children[i]->numberOfNodes();
	}
	
	return nodes;
}

int fi_tree_pointer_compare(const FITree* is1, const FITree* is2) {
	return (*is1 < *is2);
}

int fi_tree_pointer_support_compare(const FITree* is1, const FITree* is2) {
	if (is1->support == is2->support) {
		return (is1 > is2);
	}
	
	return (is1->support > is2->support);
}

// Expands the FITree from a closed itemset representation to a
// all frequent itemsets representation
// This one is still buggy....
void FITree::expand() {
	if (parent != NULL && items.size() >= 2) {
		Item* first_item = items[0];
		vector<FITree*> new_children;

		for (int i = 1; i < items.size(); i++) {
			FITree* new_child = new FITree(*this);
			new_child->clearItems();
			for (int j = i; j < items.size(); j++) {
				new_child->addItem(items[j]);
			}
			new_children.push_back(new_child);
		}
		
		clearItems();
		addItem(first_item);
		
		for (int i = 0; i < new_children.size(); i++) {
			addChild(new_children[i]);
		}
		sort(children.begin(), children.end(), fi_tree_pointer_compare);
	}
	
	for (int i = 0; i < children.size(); i++) {
		children[i]->expand();
	}
}

void FITree::outputTree(ostream& os) {
	FITree* p = parent;
	while (p != NULL) {
		os << "\t";
		p = p->parent;
	}
	
	for (int i = 0; i < items.size(); i++) {
		os << *(items[i]) << ", ";
	}
	os << "s=(" << support << ") " << "l,u=[" << lower_bound << "," << upper_bound << "] " << "tidhash=(" << tidhash << ")" << std::endl;
	
	for (int i = 0; i < children.size(); i++) {
		children[i]->outputTree(os);
	}
}

// Outputs the itemsets to a stream
void FITree::outputItemsets(ostream& os) {
	vector<Item*>* itemset = new vector<Item*>;
	if (parent != NULL) {
		itemset->push_back(items[0]);
		outputItemsets(itemset, os, 1);
		itemset->pop_back();
	} else {
		outputItemsets(itemset, os, 0);
	}
	delete itemset;
}

void FITree::outputItemsets(vector<Item*>* itemset, ostream& os) {
	if (parent != NULL) {
		itemset->push_back(items[0]);
		outputItemsets(itemset, os, 1);
		itemset->pop_back();
	} else {
		outputItemsets(itemset, os, 0);
	}
}

void FITree::outputItemsets(vector<Item*>* itemset, ostream& os, int from_index) {
	if (from_index >= items.size()) {
		for (int i = 0; i < itemset->size(); i++) {
			os << *((*itemset)[i]) << " ";
		}
		
		if (this->cover == NULL) {
			os << "(" << support << ") " << "(" << (double)(support)/(double)(this->root()->support) << ")" << std::endl;
		} else {
			Cover* this_cover = getCover();

			vector<int> tids;
			for (int i = 0; i < this_cover->size; i++) {
				tids.push_back(this_cover->tid_table[i]);
			}
			sort(tids.begin(), tids.end());
			
			int context_changes = 0;
			for (int i = 0; i < tids.size(); i++) {
				if (i == 0) {
					context_changes++;
				} else if (tids[i] - 1 != tids[i-1]) {
					context_changes++;
				}
			}

			os << "(" << support << ") " << "(" << (double)(support)/(double)(this->root()->support) << ") " << "(" << support/context_changes << ")" << std::endl;
			delete this_cover;
		}
		// For each itemset in this node, also do the itemsets
		// containing the itemset.
		for (int i = 0; i < children.size(); i++) {
			children[i]->outputItemsets(itemset, os);
		}
	} else {
		// Without item in from_index
		outputItemsets(itemset, os, from_index + 1);

		// With item in from_index
		itemset->push_back(items[from_index]);
		outputItemsets(itemset, os, from_index + 1);
		itemset->pop_back();
	}		
}
