/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
     
/**
 * A frequent itemset & association rule miner - FIRM
 * @author Juho Muhonen
 * @modification 11.11.2005 Added J-measure
 * @modification 10.11.2005 Added GPL/copyright to all files
 * @modification 9.11.2005 Tweaked the test output
 * @modification 7.11.2005 Median duration for context itemsets
 * @modification 31.10.2005 Size constraint, can now define min freq as support or frequency, short usage
 * @modification 26.10.2005 Lift-constraint, MACROS
 * @modification 25.10.2005 support constraint, confidence constraint, MACROS
 * @modification 24.10.2005 php-output
 * @modification 6.10.2005 commented all
 */

#define _PATTERN_ITEMSET "itemset"
#define _PATTERN_CLOSED "closed"
#define _PATTERN_NDI "ndi"
#define _PATTERN_ARULE "arule"
#define _PATTERN_NDAR "ndar"

#define _DATASET_AV "av"
#define _DATASET_CONTEXT "context"
#define _DATASET_TRANSACTION "transaction"

#define _OUTFORMAT_TEXT "text"
#define _OUTFORMAT_HTML "html"
#define _OUTFORMAT_PHP "php"


#include <fstream>
#include <iostream>
using std::ostream;
using std::ofstream;
using std::cout;
using std::cerr;
using std::endl;
using std::flush;

#include "ParameterParser.h"

#include "PhpOutputGenerator.h"
#include "HtmlOutputGenerator.h"
#include "TextOutputGenerator.h"
#include "SupportConstraint.h"
#include "FrequencyConstraint.h"
#include "SizeConstraint.h"
#include "ConfidenceConstraint.h"
#include "JMeasureConstraint.h"
#include "LiftConstraint.h"
#include "ContextFrequencyConstraint.h"
#include "ContextAverageDurationConstraint.h"
#include "ContextMedianDurationConstraint.h"

#include "AVTree.h"
#include "ContextTree.h"
#include "FITree.h"
#include "Eclat.h"
#include "NDI.h"
#include "RuleGenerator.h"
#include "Cover.h"
#include "AssociationRule.h"

/**
 * Print the modifications to an output stream.
 * @param os The output stream.
 */
void modifications(ostream& os) {
	os << "11.11.2005             : v3.2.4: J-measure constraint" << endl;
	os << "10.11.2005             : v3.2.3: firm is now copyrighted and distributed under GPL" << endl;
	os << "9.11.2005              : v3.2.3: Tweaked the test output" << endl;
	os << "7.11.2005              : v3.2.2: Median duration for context itemsets" << endl;
	os << "31.10.2005             : v3.2.2: Size constraint, can now define min freq as support or frequency, short usage" << endl;
	os << "26.10.2005             : v3.2.2: Lift constraint" << endl;
	os << "25.10.2005             : v3.2.2: Support constraint, Confidence constraint, MACROS" << endl;
	os << "24.10.2005             : v3.2.2: Introduced Php-output (all sorting and data in one file)" << endl;
	os << "6.10.2005 - 15.10.2005 : v3.2.1: Commented all" << endl;
	os << "23.9.2005 - 28.9.2005  : v3.2.1: Introduced context frequency, average context duration" << endl;
	os << "9.9.2005 - 22.9.2005   : v3.2.0: Introduced various changes. For example closed itemset mining" << endl;
	os << "8.9.2005               : v3.1.1: Introduced the context-tree" << endl;
	os << "22.12.2004             : v3.1.0: This is the first version to succesfully implement incremental pruning" << endl;
	os << "23.11.2004             : v3.0.1" << endl;
}

/**
 * Print the usage of the program to an output stream.
 * @param os The output stream.
 */
void usage(ostream& os) {
	os << "Usage: firm [options]" << endl;
	os << endl;
	os << "Note: You must define AT LEAST support (-s) OR frequency (-f) for actual mining of itemsets." << endl;
	os << "      If both are defined then support is used for mining and frequency for pruning." << endl;
	os << endl;
	os << "General options:" << endl;
	os << "-help                         : Display full information on options" << endl;
	os << "-i <input file>               : <input file> = The name of the input file." << endl;
	os << "-out <output file>            : Set the name of the output file." << endl;
	os << "                              : <output file> = The name of the output file, default = standard output" << endl;
	os << "-sout <output file>           : Set the name of the summary output file." << endl;
	os << "                              : <output file> = The name of the summary output file, default = standard error output" << endl;
	os << "-outformat <output format>    : Set the output format of the patterns (-out). If other format than text is selected," << endl;
	os << "                              : an additional summary file will still be given in text format as specified by (-sout)." << endl;
	os << "                              : The php format supports sorting if the result file is located on a server capable of handling" << endl;
	os << "                              : php-scripts." << endl;
	os << "                              : <output format> = " << _OUTFORMAT_TEXT << ", (text format)" << endl;
	os << "                              :                or " << _OUTFORMAT_HTML << ", (html format)" << endl;
	os << "                              :                or " << _OUTFORMAT_PHP << ",  (php format)" << endl;
	os << "                              :                default = " << _OUTFORMAT_TEXT << "." << endl;
	os << "-pattern <type>               : Set the type of patterns you want to mine." << endl;
	os << "                              : <type> = " << _PATTERN_ITEMSET << ", (frequent itemsets)" << endl;
	os << "                              :       or " << _PATTERN_CLOSED << ", (frequent closed itemsets)" << endl;
	os << "                              :       or " << _PATTERN_NDI << ", (frequent non-derivable itemsets)" << endl;
	os << "                              :       or " << _PATTERN_ARULE << ", (frequent association rules)" << endl;
	os << "                              :       or " << _PATTERN_NDAR << ", (frequent non-derivable association rules)" << endl;
	os << "                              :       default = " << _PATTERN_ITEMSET << "" << endl;
	os << "-dataset <type>               : Set the type of the dataset you are using." << endl;
	os << "                              : <type> = " << _DATASET_AV << ", (attribute-value data format). See 'exampleset.firm' for example." << endl;
	os << "                              :       or " << _DATASET_TRANSACTION << ", (transaction data format). See 'exampleset_transaction.firm' for example." << endl;
	os << "                              :       or " << _DATASET_CONTEXT << ", (context data format). See 'exampleset_context.firm' for example." << endl;
	os << "                              :       default = " << _DATASET_AV << "." << endl;
	os << endl;
	os << "Itemset options (for all itemsets):" << endl;
	os << "-size <max size>              : Set the maximum size of an itemset." << endl;
	os << "                              : <max size> = [0, oo[, default = oo." << endl;
	os << "-s <min support>              : Set the minimum absolute frequency of an itemset." << endl;
	os << "                              : <min frequency> = [0, oo[, default = 0." << endl;
	os << "-f <min frequency>            : Set the minimum relative frequency of an itemset." << endl;
	os << "                              : <min frequency> = [0, 1], default = 0." << endl;
	os << "-l <min lift>                 : Set the minimum lift of an itemset (For lifts between 0 and 1 the comparison will be made to 1/lift)" << endl;
	os << "                              : <min lift> = [1, oo[, default = 1." << endl;
	os << endl;
	os << "Context itemset specific options:" << endl;
	os << "-dc <duration column>         : REQUIRED. If the dataset is in context format, you're required to give the name of the duration column." << endl;
	os << "                              : <duration_column> = The column name of the column giving time of change from the beginning in seconds." << endl;
	os << "-avgd <min duration>          : Set the minimum average duration of a context (in hours)." << endl;
	os << "                              : <min duration> = [0, oo[, default = 0." << endl;
	os << "-medd <min duration>          : Set the minimum median duration of a context (in hours)." << endl;
	os << "                              : <min duration> = [0, oo[, default = 0." << endl;
	os << endl;
	os << "Association Rule options (for all association rules):" << endl;
	os << "-size <max size>              : Set the maximum size of an association rule." << endl;
	os << "                              : <max size> = [0, oo[, default = oo." << endl;
	os << "-s <min support>              : Set the minimum absolute frequency of an association rule." << endl;
	os << "                              : <min frequency> = [0, oo[, default = 0." << endl;
	os << "-f <min frequency>            : Set the minimum relative frequency of an association rule." << endl;
	os << "                              : <min frequency> = [0, 1], default = 0." << endl;
	os << "-c <min confidence>           : Set the minimum confidence of an association rule." << endl;
	os << "                              : <min confidence> = [0, 1], default = 0." << endl;
	os << "-l <min lift>                 : Set the minimum lift of an association rule. (For lifts between 0 and 1 the comparison will be made to 1/lift)" << endl;
	os << "                              : <min lift> = [1, oo[, default = 1." << endl;
	os << "-j <min J-measure>            : Set the minimum J-measure of an association rule." << endl;
	os << "                              : <min lift> = [0, 1], default = 0." << endl;
	os << endl;
	os << "Non-derivable Association Rule options:" << endl;
	os << "-b <min difference of bounds> : Set the maximum absolute difference of the upper and lower bound of confidence of an association rule." << endl;
	os << "                              : <min difference of bounds> = [0, 1], default = 0" << endl;
}

/**
 * Print the short usage of the program to an output stream.
 * @param os The output stream.
 */
void shortUsage(ostream& os) {
	os << "Usage: firm [options]" << endl;
	os << endl;
	os << "Note: You must define AT LEAST support (-s) OR frequency (-f) for actual mining of itemsets." << endl;
	os << "      If both are defined then support is used for mining and frequency for pruning." << endl;
	os << endl;
	os << "General options:" << endl;
	os << "-help                         : Display full information on options" << endl;
	os << "-i <input file>               : <input file> = The name of the input file." << endl;
	os << "-out <output file>            : Set the name of the output file." << endl;
	os << "-sout <output file>           : Set the name of the summary output file." << endl;
	os << "-outformat <output format>    : Set the output format of the patterns (-out). If other format than text is selected," << endl;
	os << "                              : an additional summary file will still be given in text format as specified by (-sout)." << endl;
	os << "                              : The php format supports sorting if the result file is located on a server capable of handling" << endl;
	os << "                              : php-scripts." << endl;
	os << "-pattern <type>               : Set the type of patterns you want to mine." << endl;
	os << "-dataset <type>               : Set the type of the dataset you are using." << endl;
	os << endl;
	os << "Itemset options (for all itemsets):" << endl;
	os << "-size <max size>              : Set the maximum size of an itemset." << endl;
	os << "-s <min support>              : Set the minimum absolute frequency of an itemset." << endl;
	os << "-f <min frequency>            : Set the minimum relative frequency of an itemset." << endl;
	os << "-l <min lift>                 : Set the minimum lift of an itemset (For lifts between 0 and 1 the comparison will be made to 1/lift)" << endl;
	os << endl;
	os << "Context itemset specific options:" << endl;
	os << "-dc <duration column>         : REQUIRED. If the dataset is in context format, you're required to give the name of the duration column." << endl;
	os << "-avgd <min duration>          : Set the minimum average duration of a context (in hours)." << endl;
	os << "-medd <min duration>          : Set the minimum median duration of a context (in hours)." << endl;
	os << endl;
	os << "Association Rule options (for all association rules):" << endl;
	os << "-size <max size>              : Set the maximum size of an association rule." << endl;
	os << "-s <min support>              : Set the minimum absolute frequency of an association rule." << endl;
	os << "-f <min frequency>            : Set the minimum relative frequency of an association rule." << endl;
	os << "-c <min confidence>           : Set the minimum confidence of an association rule." << endl;
	os << "-l <min lift>                 : Set the minimum lift of an association rule. (For lifts between 0 and 1 the comparison will be made to 1/lift)" << endl;
	os << "-j <min J-measure>            : Set the minimum J-measure of an association rule." << endl;
	os << endl;
	os << "Non-derivable Association Rule options:" << endl;
	os << "-b <min difference of bounds> : Set the maximum absolute difference of the upper and lower bound of confidence of an association rule." << endl;
}

/**
 * Print the version of the program to an output stream.
 * @param os The output stream.
 */
void version(ostream& os) {
	os << "FIRM - Frequent Itemset & Rule Mining Software" << endl;
	os << "Copyright (c) Juho Muhonen 2004-2005"<< endl;
	os << "Version: 3.2.4" << endl;
	os << endl;
}

/**
 * The main program.
 * @param arcg The number of parameters
 * @param argc An array of char arrays of parameters.
 */
int main(int argc, char* argv[]) {
	// If there are no parameters print the usage.
	if (argc == 1) {
		version(cout);
		shortUsage(cout);
		exit(0);
	}

	// Parse the parameters.
	ParameterParser parameters(argc, argv);
	parameters.addParameter("column_separator", "\t");
	parameters.addParameter("transaction_separator", "\n");

	if (parameters.hasParameter("help")) {
		version(cout);
		usage(cout);
		exit(0);
	} 

	if ((!parameters.hasParameter("s") && !parameters.hasParameter("f")) || (parameters.isFlagParameter("s") && !parameters.isFlagParameter("f"))) {
		cerr << "Support (-s) or Freqeuency (-f) must be specified" << endl;
		shortUsage(cerr);
		exit(0);
	} 
	
	if (!parameters.hasParameter("i")) {
		cerr << "Input file (-i) must must be specified" << endl;
		shortUsage(cerr);
		exit(0);
	}
	
	string pattern = _PATTERN_ITEMSET;
	if (parameters.hasParameter("pattern")) {
		pattern = parameters.getParameter("pattern");;
		if (pattern != _PATTERN_ITEMSET && pattern != _PATTERN_CLOSED && pattern != _PATTERN_NDI && pattern != _PATTERN_ARULE && pattern != _PATTERN_NDAR) {
			cerr << "Invalid pattern type (-pattern)" << endl;
			shortUsage(cerr);
			exit(0);
		}
	}

	string dataset = _DATASET_AV;
	if (parameters.hasParameter("dataset")) {
		dataset = parameters.getParameter("dataset");
		if (dataset != _DATASET_AV && dataset != _DATASET_TRANSACTION && dataset != _DATASET_CONTEXT){
			cerr << "Invalid dataset type (-dataset)" << endl;
			shortUsage(cerr);
			exit(0);
		}
	}

	if (dataset == _DATASET_CONTEXT && !parameters.hasParameter("dc")) {
		cerr << "Context duration column has to be specified for context datasets (-dc)" << endl;
		shortUsage(cerr);
		exit(0);
	} 

	ostream* sout = &cerr;
	ofstream* file_sout = NULL;
	if (parameters.hasParameter("sout")) {
		file_sout = new ofstream(parameters.getParameter("sout").c_str());
		sout = file_sout;
	}

	string output_format = _OUTFORMAT_TEXT;
	OutputGenerator* outputter;
	if (parameters.hasParameter("outformat")) {
		output_format = parameters.getParameter("outformat");
	}
	if (output_format == _OUTFORMAT_TEXT) {
		outputter = new TextOutputGenerator();
	} else if (output_format == _OUTFORMAT_HTML) {
		outputter = new HtmlOutputGenerator();
	} else if (output_format == _OUTFORMAT_PHP) {
		outputter = new PhpOutputGenerator();
	} else {
		cerr << "Invalid output format (-outformat)" << endl;
		usage(cerr);
		exit(0);
	}

	if (parameters.hasParameter("out")) {
		outputter->setOutputName(parameters.getParameter("out"));
	}

	clock_t start;
	version(*sout);
	*sout << "Running with parameters:" << endl << parameters;

	start = clock();
	*sout << "Reading the dataset " << flush;
	AVTree* av_data;
	ContextTree* context_data;
	int min_support;
	if (dataset == _DATASET_AV) {
		av_data = new AVTree(parameters.getParameter("i"), parameters.getParameter("column_separator").at(0), parameters.getParameter("transaction_separator").at(0));
		if (parameters.hasParameter("s") && !parameters.isFlagParameter("s")) {
			min_support = atoi(parameters.getParameter("s").c_str());
		} else {
			min_support = (int)(av_data->getTransactions() * atof(parameters.getParameter("f").c_str()));
		}
	} else if (dataset == _DATASET_TRANSACTION) {
		av_data = new AVTree(parameters.getParameter("i"), parameters.getParameter("column_separator").at(0), parameters.getParameter("transaction_separator").at(0), true);
		if (parameters.hasParameter("s") && !parameters.isFlagParameter("s")) {
			min_support = atoi(parameters.getParameter("s").c_str());
		} else {
			min_support = (int)(av_data->getTransactions() * atof(parameters.getParameter("f").c_str()));
		}
	} else if (dataset == _DATASET_CONTEXT) {
		context_data = new ContextTree(parameters.getParameter("i"), parameters.getParameter("dc"), parameters.getParameter("column_separator").at(0), parameters.getParameter("transaction_separator").at(0));
		if (parameters.hasParameter("s") && !parameters.isFlagParameter("s")) {
			min_support = atoi(parameters.getParameter("s").c_str());
		} else {
			min_support = (int)(context_data->getTransactions() * atof(parameters.getParameter("f").c_str()));
		}
	}
	*sout << "[" << (clock()-start)/double(CLOCKS_PER_SEC) << "s]" << endl;

	if (dataset == _DATASET_AV || dataset == _DATASET_TRANSACTION) {
		if (pattern == _PATTERN_ITEMSET || pattern == _PATTERN_NDI || pattern == _PATTERN_CLOSED) {
			if (parameters.hasParameter("size")) {
				if (parameters.isFlagParameter("size")) {
					outputter->addSetConstraint(new SizeConstraint());
				} else {
					outputter->addSetConstraint(new SizeConstraint(atoi(parameters.getParameter("size").c_str())));
				}
			}
			if (parameters.hasParameter("s")) {
				if (parameters.isFlagParameter("s")) {
					outputter->addSetConstraint(new SupportConstraint());
				} else {
					outputter->addSetConstraint(new SupportConstraint(atoi(parameters.getParameter("s").c_str())));
				}
			}
			if (parameters.hasParameter("f")) {
				if (parameters.isFlagParameter("f")) {
					outputter->addSetConstraint(new FrequencyConstraint());
				} else {
					outputter->addSetConstraint(new FrequencyConstraint(atof(parameters.getParameter("f").c_str())));
				}
			}
			if (parameters.hasParameter("l")) {
				if (parameters.isFlagParameter("l")) {
					outputter->addSetConstraint(new LiftConstraint());
				} else {
					outputter->addSetConstraint(new LiftConstraint(atof(parameters.getParameter("l").c_str())));
				}
			}
		} else if (pattern == _PATTERN_ARULE) {
			if (parameters.hasParameter("size")) {
				if (parameters.isFlagParameter("size")) {
					outputter->addRuleConstraint(new SizeConstraint());
				} else {
					outputter->addRuleConstraint(new SizeConstraint(atoi(parameters.getParameter("size").c_str())));
				}
			}
			if (parameters.hasParameter("s")) {
				if (parameters.isFlagParameter("s")) {
					outputter->addRuleConstraint(new SupportConstraint());
				} else {
					outputter->addRuleConstraint(new SupportConstraint(atoi(parameters.getParameter("s").c_str())));
				}
			}
			if (parameters.hasParameter("f")) {
				if (parameters.isFlagParameter("f")) {
					outputter->addRuleConstraint(new FrequencyConstraint());
				} else {
					outputter->addRuleConstraint(new FrequencyConstraint(atof(parameters.getParameter("f").c_str())));
				}
			}
			if (parameters.hasParameter("c")) {
				if (parameters.isFlagParameter("c")) {
					outputter->addRuleConstraint(new ConfidenceConstraint());
				} else {
					outputter->addRuleConstraint(new ConfidenceConstraint(atof(parameters.getParameter("c").c_str())));
				}
			}
			if (parameters.hasParameter("l")) {
				if (parameters.isFlagParameter("l")) {
					outputter->addRuleConstraint(new LiftConstraint());
				} else {
					outputter->addRuleConstraint(new LiftConstraint(atof(parameters.getParameter("l").c_str())));
				}
			}
			if (parameters.hasParameter("j")) {
				if (parameters.isFlagParameter("j")) {
					outputter->addRuleConstraint(new JMeasureConstraint());
				} else {
					outputter->addRuleConstraint(new JMeasureConstraint(atof(parameters.getParameter("j").c_str())));
				}
			}
		}
	} else if (dataset == _DATASET_CONTEXT) {
		if (pattern == _PATTERN_ITEMSET || pattern == _PATTERN_NDI || pattern == _PATTERN_CLOSED) {
			if (parameters.hasParameter("size")) {
				if (parameters.isFlagParameter("size")) {
					outputter->addSetConstraint(new SizeConstraint());
				} else {
					outputter->addSetConstraint(new SizeConstraint(atoi(parameters.getParameter("size").c_str())));
				}
			}
			if (parameters.hasParameter("s")) {
				if (parameters.isFlagParameter("s")) {
					outputter->addSetConstraint(new SupportConstraint());
				} else {
					outputter->addSetConstraint(new SupportConstraint(atoi(parameters.getParameter("s").c_str())));
				}
			}
			if (parameters.hasParameter("f")) {
				if (parameters.isFlagParameter("f")) {
					outputter->addSetConstraint(new FrequencyConstraint());
				} else {
					outputter->addSetConstraint(new FrequencyConstraint(atof(parameters.getParameter("f").c_str())));
				}
			}
			if (parameters.hasParameter("l")) {
				if (parameters.isFlagParameter("l")) {
					outputter->addSetConstraint(new LiftConstraint());
				} else {
					outputter->addSetConstraint(new LiftConstraint(atof(parameters.getParameter("l").c_str())));
				}
			}
			if (parameters.hasParameter("avgd")) {
				ContextAverageDurationConstraint* cfc = new ContextAverageDurationConstraint();
				if (!parameters.isFlagParameter("avgd")) {
					cfc->setMinDuration(atof(parameters.getParameter("avgd").c_str()));
				}
				outputter->addSetConstraint(cfc);
			}
			if (parameters.hasParameter("medd")) {
				ContextMedianDurationConstraint* cfc = new ContextMedianDurationConstraint();
				cfc->setContextTree(context_data);
				if (!parameters.isFlagParameter("medd")) {
					cfc->setMinDuration(atof(parameters.getParameter("avgd").c_str()));
				}
				outputter->addSetConstraint(cfc);
			}
		}
	}
	
	if (dataset == _DATASET_AV || dataset == _DATASET_TRANSACTION) {
		if (pattern == _PATTERN_ITEMSET) {
			*sout << "Generating all frequent itemsets " << flush;
			Eclat eclat(true, true);
			start = clock();
			FITree* itemsets = eclat.generateFrequentSets(av_data, min_support);
			*sout << "[" << (clock()-start)/double(CLOCKS_PER_SEC) << "s]" << endl << flush;

			*sout << "Generating output " << flush;
			start = clock();
			outputter->outputSets(itemsets);
			*sout << "[" << (clock()-start)/double(CLOCKS_PER_SEC) << "s]" << endl << flush;

			*sout << "Total number of frequent itemsets: " << itemsets->numberOfFrequentSets() << endl;
			delete itemsets;
		} else if (pattern == _PATTERN_NDI) {
			*sout << "Generating non-derivable itemsets " << flush;
			NDI ndi;
			start = clock();
			FITree* ndis = ndi.generateNDI(av_data, min_support);
			*sout << "[" << (clock()-start)/double(CLOCKS_PER_SEC) << "s]" << endl << flush;

			*sout << "Generating output " << flush;
			start = clock();
			outputter->outputSets(ndis);
			*sout << "[" << (clock()-start)/double(CLOCKS_PER_SEC) << "s]" << endl << flush;

			*sout << "Total number of non-derivable itemsets: " << ndis->numberOfFrequentSets() << endl;
			delete ndis;
		} else if (pattern == _PATTERN_CLOSED) {
			*sout << "Generating closed frequent itemsets " << flush;
			Eclat eclat;
			start = clock();
			FITree* closeds = eclat.generateFrequentClosedSets(av_data, min_support);
			*sout << "[" << (clock()-start)/double(CLOCKS_PER_SEC) << "s]" << endl << flush;

			*sout << "Generating output " << flush;
			start = clock();
			outputter->outputSets(closeds);
			*sout << "[" << (clock()-start)/double(CLOCKS_PER_SEC) << "s]" << endl << flush;

			*sout << "Total number of closed itemsets: " << closeds->numberOfNodes() << endl;
			delete closeds;
		} else if (pattern == _PATTERN_NDAR) {
			*sout << "Generating all frequent itemsets " << flush;
			Eclat eclat(true, true);
			start = clock();
			FITree* itemsets = eclat.generateFrequentSets(av_data, min_support);
			*sout << "[" << (clock()-start)/double(CLOCKS_PER_SEC) << "s]" << endl << flush;
			*sout << "Total number of frequent itemsets: " << itemsets->numberOfFrequentSets() << endl;
			*sout << "Generating non-derivable association rules " << flush;

			RuleGenerator generator;
			if (parameters.hasParameter("c")) {
				generator.setMinConfidence(atof(parameters.getParameter("c").c_str()));
			} else {
				generator.setMinConfidence(0);
			}

			if (parameters.hasParameter("b")) {
				generator.setMinDifferenceOfBounds(-1);
			} else {
				generator.setMinDifferenceOfBounds(atof(parameters.getParameter("b").c_str()));
			}

			start = clock();
			generator.generateRules(itemsets, *sout);
			*sout << "[" << (clock()-start)/double(CLOCKS_PER_SEC) << "s]" << endl << flush;

			*sout << "Total number of association rules: " << generator.rulesGenerated() << endl;
			*sout << "Rules that are confident         : " << generator.rulesConfident() << endl;
			*sout << "Non pruned association rules     : " << generator.rulesOutput() << endl;
			*sout << "Association rules with zero DoB  : " << generator.rulesWithZeroDob() << endl;
		} else if (pattern == _PATTERN_ARULE) {
			*sout << "Generating all frequent itemsets " << flush;
			Eclat eclat(true, true);
			start = clock();
			FITree* itemsets = eclat.generateFrequentSets(av_data, min_support);
			*sout << "[" << (clock()-start)/double(CLOCKS_PER_SEC) << "s]" << endl << flush;
			*sout << "Total number of frequent itemsets: " << itemsets->numberOfFrequentSets() << endl;
			start = clock();

			*sout << "Generating association rules and output" << flush;
			outputter->outputRules(itemsets);
			*sout << "[" << (clock()-start)/double(CLOCKS_PER_SEC) << "s]" << endl << flush;

			delete itemsets;
		}
	} else if (dataset == _DATASET_CONTEXT) {
		if (pattern == _PATTERN_ITEMSET) {
			*sout << "Generating context frequent itemsets " << flush;
			Eclat eclat(true,true);
			start = clock();
			FITree* itemsets = eclat.generateFrequentSets(context_data, min_support);
			*sout << "[" << (clock()-start)/double(CLOCKS_PER_SEC) << "s]" << endl << flush;

			*sout << "Generating output " << flush;
			start = clock();
			outputter->outputSets(itemsets);
			*sout << "[" << (clock()-start)/double(CLOCKS_PER_SEC) << "s]" << endl << flush;

			*sout << "Total number of context frequent itemsets: " << itemsets->numberOfFrequentSets() << endl;
			delete itemsets;
		} else if (pattern == _PATTERN_NDI) {
			*sout << "Generating non-derivable context itemsets " << flush;
			NDI ndi;
			start = clock();
			FITree* ndis = ndi.generateNDI(context_data, min_support);
			*sout << "[" << (clock()-start)/double(CLOCKS_PER_SEC) << "s]" << endl << flush;

			*sout << "Generating output " << flush;
			start = clock();
			outputter->outputSets(ndis);
			*sout << "[" << (clock()-start)/double(CLOCKS_PER_SEC) << "s]" << endl << flush;

			*sout << "Total number of non-derivable itemsets: " << ndis->numberOfNodes() << endl;
			delete ndis;
		} else if (pattern == _PATTERN_CLOSED) {
			*sout << "Generating frequent closed context itemsets " << flush;
			Eclat eclat_closed;
			start = clock();
			FITree* closeds = eclat_closed.generateFrequentClosedContextSets(context_data, min_support);
			*sout << "[" << (clock()-start)/double(CLOCKS_PER_SEC) << "s]" << endl << flush;

			*sout << "Generating output " << flush;
			start = clock();
			outputter->outputSets(closeds);
			*sout << "[" << (clock()-start)/double(CLOCKS_PER_SEC) << "s]" << endl << flush;

			*sout << "Total number of closed itemsets: " << closeds->numberOfNodes() << endl;
			delete closeds;
		} else {
			*sout << "Sorry, only generation of context itemsets is supported at the moment" << endl;
		}
	}
	
	delete outputter;
}

