/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "FrequencyConstraint.h"
#include "FITree.h"
#include "AssociationRule.h"

/**
 * Create a frequency constraint with minimum frequency of 0
 */
FrequencyConstraint::FrequencyConstraint() {
	last_frequency = 0;
	min_frequency = 0;
}

/**
 * Create a frequency constraint with a minimum frequency.
 * @param d The minimum frequency.
 */
FrequencyConstraint::FrequencyConstraint(double d) {
	last_frequency = 0;
	min_frequency = d;
}

/**
 * Set the minimum frequency of this constraint.
 * @param d The minimum frequency.
 */
void FrequencyConstraint::setMinFrequency(double d) {
	min_frequency = d;
}

/**
 * Returns the result of last evaluation.
 * @return The frequency of the pattern evaluated last.
 */
double FrequencyConstraint::lastResult() {
	return last_frequency;
}

/**
 * Value an itemset.
 * @param fset The itemset.
 * @return The frequency of the itemset.
 */
double FrequencyConstraint::value(FITree* fset) {
	last_frequency = (double)fset->support / (double)fset->root()->support;
	return last_frequency;
}

/**
 * Value an association rule.
 * @param arule The association rule.
 * @return The frequency of the association rule.
 */
double FrequencyConstraint::value(AssociationRule* arule) {
	last_frequency = (double)arule->itemset->support / (double)arule->itemset->root()->support;
	return last_frequency;
}

/**
 * Validate an itemset.
 * @param fset The itemset.
 * @return True if the frequency of the itemset is higher than the minimum
 * frequency of this constraint.
 */
bool FrequencyConstraint::validate(FITree* fset) {
	if (value(fset) >= min_frequency) {
		return true;
	}
	
	return false;
}

/**
 * Validate an association rule.
 * @param arule The association rule.
 * @return True if the frequency of the rule is higher than the minimum
 * frequency of this constraint.
 */
bool FrequencyConstraint::validate(AssociationRule* arule) {
	if (value(arule) >= min_frequency) {
		return true;
	}
	
	return false;
}

/**
 * Returns the minimum frequency.
 * @return The minimum frequency.
 */
double FrequencyConstraint::constraintLimit() const {
	return min_frequency;
}

/**
 * Returns the name of this constraint.
 * @return The name.
 */
string FrequencyConstraint::name() const {
	return "Frequency";
}

/**
 * Returns the abbreviation of this constraint.
 * @return The abbreviation.
 */
string FrequencyConstraint::abbreviation() const {
	return "f";
}

/**
 * Returns the description of this constraint.
 * @return The description.
 */
string FrequencyConstraint::description() const {
	return "Relative frequency is the support of the pattern divided by the support of the dataset.";
}
