/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "HtmlOutputGenerator.h"
#include "SetConstraint.h"
#include "RuleConstraint.h"

#include "Item.h"
#include "FITree.h"
#include "AssociationRule.h"

#include <iostream>
using std::ostream;

#include <queue>
using std::queue;

/**
 * Creates a new html output generator.
 */
HtmlOutputGenerator::HtmlOutputGenerator() {
	header_line = true;
	output_name = "";
	output_stream = NULL;
	header_stream = NULL;
	frame_stream = NULL;
}

/**
 * Prints a double in color (html) using default values.
 * @param d The double
 */
void HtmlOutputGenerator::doubleInColor(double d) {
	doubleInColor(d, 1.0, -1.0, 0.0, 1, true);
}

/**
 * Outputs a double in color (html) using given upper and lower bound, zero limit, spacing and increasing
 * @param d The double to be printed.
 * @param upper_b The upper bound of the double (closer to the upper bound the double is, the greener the color)
 * @param upper_b The lower bound of the double (closer to the lower bound the double is, more red the color)
 * @param negative_l The limit where the color is white.
 * @param spacing
 * @param increasing Tells whether low or high values should be green.
 */
void HtmlOutputGenerator::doubleInColor(double d, double upper_b, double lower_b, double negative_l, int spacing, bool increasing) {
	double upper_bound;
	double lower_bound;
	double negative_limit;
	double value;
	
	if (increasing) {
		upper_bound = upper_b;
		lower_bound = lower_b;
		negative_limit = negative_l;
		value = d;
	} else {
		upper_bound = -lower_b;
		lower_bound = -upper_b;
		negative_limit = -negative_l;
		value = -d;
	}

	if (value > upper_bound) {
		value = upper_bound;
	}
	
	//std::cerr << "double:" << value << " upper_bound:" << upper_bound << " lower_bound:" << lower_bound << " negative_limit:" << negative_limit;
	if (d >= negative_limit) {
		int int_result = static_cast<int>(((value - negative_limit) / (upper_bound - negative_limit)) * 255 / spacing) * spacing;;
		//std::cerr << " int_result: " << int_result << std::endl;
		*output_stream << "<td bgcolor=\"#";
		*output_stream << "0123456789ABCDEF"[static_cast<int>((255 - int_result) / 16)];
		*output_stream << "0123456789ABCDEF"[(255 - int_result) % 16];
		*output_stream << "FF";
		*output_stream << "0123456789ABCDEF"[static_cast<int>((255 - int_result) / 16)];
		*output_stream << "0123456789ABCDEF"[(255 - int_result) % 16] << "\">" << d << "</td>";
	} else {
		int int_result = static_cast<int>(((value - negative_limit) / (negative_limit - lower_bound)) * 255 / spacing) * spacing;;
		//cerr << "int_result:" << int_result << "\n";
		//std::cerr << " int_result: " << int_result << std::endl;
		*output_stream << "<td bgcolor=\"#";
		*output_stream << "FF";
		*output_stream << "0123456789ABCDEF"[static_cast<int>((255 + int_result) / 16)];
		*output_stream << "0123456789ABCDEF"[(255 + int_result) % 16];
		*output_stream << "0123456789ABCDEF"[static_cast<int>((255 + int_result) / 16)];
		*output_stream << "0123456789ABCDEF"[(255 + int_result) % 16] << "\">" << d << "</td>";
	}
}


/**
 * Outputs the frame.
 */
void HtmlOutputGenerator::outputFrame() {		
	*frame_stream << "<html>";

	*frame_stream << "<frameset rows=\"200,*\">" << std::endl;

	*frame_stream << "<frame src=\"header_" << output_name << "\"" << ">" << std::endl;
	*frame_stream << "</frame>" << std::endl;

	*frame_stream << "<frame name=\"rules\" src=\"" << output_name << "\"" << ">" << std::endl;
	*frame_stream << "</frame>" << std::endl;

	*frame_stream << "</frameset>" << std::endl;
	*frame_stream << "</html>";
}

/**
 * Outputs the header for association rule output.
 */
void HtmlOutputGenerator::outputRuleHeader() {
	*header_stream << "<html>";
	*header_stream << "<body>";
	*header_stream << "<table border=\"1\">" << std::endl;
	*header_stream << "<tr>";
	*header_stream << "<td><a href=\"cgi-bin/sort_by_column.pl?rule_file=" << output_name << "&column=0&descending=1\" target=\"rules\"><b>" << "Condition" << "</b></a></td>";
	*header_stream << "<td><b>" << "->" << "</b></td>";
	*header_stream << "<td><a href=\"cgi-bin/sort_by_column.pl?rule_file=" << output_name << "&column=1&descending=1\" target=\"rules\"><b>" << "Consequent" << "</b></a></td>";
	int column_number = 2;
		
	for (vector<RuleConstraint*>::iterator cons_iter = rule_constraints.begin(); cons_iter != rule_constraints.end(); cons_iter++) { 
		if ((*cons_iter)->name() != "") {
			*header_stream << "<td><a href=\"cgi-bin/sort_by_column.pl?rule_file=" << output_name << "&column=" << column_number << "&descending=" << (*cons_iter)->increasing() << "\" target=\"rules\"><b>" << (*cons_iter)->name() << " (" << (*cons_iter)->abbreviation() << ")" << "</b></a></td>";
			column_number++;
		}
	}
	*header_stream << "</tr>" << std::endl;

	*header_stream << "<tr>";
	*header_stream << "<td><b>" << "" << "</b></td>";
	*header_stream << "<td><b>" << "" << "</b></td>";
	*header_stream << "<td><b>" << "" << "</b></td>";

	for (vector<RuleConstraint*>::iterator cons_iter = rule_constraints.begin(); cons_iter != rule_constraints.end(); cons_iter++) { 
		if ((*cons_iter)->name() != "") {
			*header_stream << "<td><b>" << (*cons_iter)->description() << "</b></td>";
		}
	}
	*header_stream << "</tr></table>" << std::endl;

	*header_stream << "</table>";
	*header_stream << "</body>";
	*header_stream << "</html>";
}

/**
 * Outputs the header for itemset output.
 */
void HtmlOutputGenerator::outputItemsetHeader() {
	*header_stream << "<html>";
	*header_stream << "<body>";
	*header_stream << "<table border=\"1\">" << std::endl;
	*header_stream << "<tr>";
	*header_stream << "<td><a href=\"cgi-bin/sort_by_column.pl?rule_file=" << output_name << "&column=0&descending=1\" target=\"rules\"><b>" << "Itemset" << "</b></a></td>";
	int column_number = 1;
		
	for (int i = 0; i < set_constraints.size(); i++) {
		*header_stream << "<td><a href=\"cgi-bin/sort_by_column.pl?rule_file=" << output_name << "&column=" << column_number << "&descending=" << set_constraints[i]->increasing() << "\" target=\"rules\"><b>" << set_constraints[i]->name() << " (" << set_constraints[i]->abbreviation() << ")" << "</b></a></td>";
		column_number++;
	}

	*header_stream << "</tr>" << std::endl;

	*header_stream << "<tr>";
	*header_stream << "<td><b></b></td>";

	for (int i = 0; i < set_constraints.size(); i++) {
		*header_stream << "<td><b>" << set_constraints[i]->description() << "</b></td>";
	}

	*header_stream << "</tr>" << std::endl;
	*header_stream << "</table>";
	*header_stream << "</body>";
	*header_stream << "</html>";
}

/**
 * Set the output name.
 * @param n The name.
 */
void HtmlOutputGenerator::setOutputName(string n) {
	assert(n != "");

	output_name = n;
	
	if (output_stream != NULL) {
		output_stream->close();
		delete output_stream;
	}	

	if (header_stream != NULL) {
		header_stream->close();
		delete header_stream;
	}	

	if (frame_stream != NULL) {
		frame_stream->close();
		delete frame_stream;
	}	

	output_stream = new ofstream(output_name.c_str());	
	
	string header_file_name("header_");
	header_file_name.append(output_name);
	header_stream = new ofstream(header_file_name.c_str());
	
	string frame_file_name("frame_");
	frame_file_name.append(output_name);
	frame_stream = new ofstream(frame_file_name.c_str());

	output_stream->setf(std::ios::fixed);
	output_stream->precision(3);

	frame_stream->setf(std::ios::fixed);
	frame_stream->precision(3);

	frame_stream->setf(std::ios::fixed);
	frame_stream->precision(3);

	outputFrame();
}

/**
 * Add a set constraint to this output generator.
 * @param c The set constraint.
 */
void HtmlOutputGenerator::addSetConstraint(SetConstraint* c) {
	assert(c != NULL);
	set_constraints.push_back(c);
}

/**
 * Add a rule constraint to this output generator.
 * @param c The rule constraint.
 */
void HtmlOutputGenerator::addRuleConstraint(RuleConstraint* c) {
	assert(c != NULL);
	rule_constraints.push_back(c);
}

/**
 * Removes a constraint by it's name.
 * @param s The constraint name.
 */
void HtmlOutputGenerator::removeConstraint(string s) {
	for (int i = 0; i < set_constraints.size(); i++) {
		if (set_constraints[i]->name() == s) {
			set_constraints.erase(set_constraints.begin() + i);
		}
	}

	for (int i = 0; i < rule_constraints.size(); i++) {
		if (rule_constraints[i]->name() == s) {
			rule_constraints.erase(rule_constraints.begin() + i);
		}
	}
}

/**
 * Outputs an itemset.
 * @param set The itemset.
 */
void HtmlOutputGenerator::outputSet(FITree* set) {
	assert(set != NULL);

	SetConstraint* constraint;
	
	bool valid = true;
	
	for (int i = 0; i < set_constraints.size() && valid; i++) {
		constraint = set_constraints[i];
		if (!constraint->validate(set)) {
			valid = false;
		}
	}

	if (valid) {

		*output_stream << "<tr>";
		*output_stream << "<td>";
	
		FITree* iter = set;
		while (iter != NULL) {
			for (int i = 0; i < iter->items.size(); i++) {
				*output_stream << *(iter->items[i]) << "<br>";
			}
			iter = iter->parent;
		}
	
		*output_stream << "</td>";
	
		for (int i = 0; i < set_constraints.size(); i++) {
			constraint = set_constraints[i];
			*output_stream << "<td><b>" << constraint->abbreviation() << "=</b></td>";
			doubleInColor(constraint->value(set), constraint->upperBound(), constraint->lowerBound(), constraint->zeroLimit(), 1, constraint->increasing());
		}
	
		*output_stream << "</tr>" << std::endl;
	}
	
	for (int i = 0; i < set->children.size(); i++) {
		outputSet(set->children[i]);
	}
}

/**
 * Outputs all itemsets strating from given frequent itemset root.
 * @param set The root of the tree.
 */
void HtmlOutputGenerator::outputSets(FITree* set) {
	assert(set != NULL);

	outputItemsetHeader();
	
	*output_stream << "<html>";
	*output_stream << "<body>";
	*output_stream << "<table border=\"1\">" << std::endl;

	outputSet(set);
	
	*output_stream << "</table>";
	*output_stream << "</body>";
	*output_stream << "</html>";
}

/**
 * Outputs an association rule.
 * @param rule The association rule.
 */
void HtmlOutputGenerator::outputRule(AssociationRule* rule) {
	assert(rule != NULL);

	RuleConstraint* constraint;
	
	bool valid = true;
	
	for (int i = 0; i < rule_constraints.size() && valid; i++) {
		constraint = rule_constraints[i];
		if (!constraint->validate(rule)) {
			valid = false;
		}
	}

	if (valid) {

		*output_stream << "<tr>";
		*output_stream << "<td>";
	
		FITree* iter = rule->condition;
		while (iter != NULL) {
			for (int i = 0; i < iter->items.size(); i++) {
				*output_stream << *(iter->items[i]) << "<br>";
			}
			iter = iter->parent;
		}
	
		*output_stream << "</td><td>-></td>";
		*output_stream << "<td>";
	
		iter = rule->consequent;
		while (iter != NULL) {
			for (int i = 0; i < iter->items.size(); i++) {
				*output_stream << *(iter->items[i]) << "<br>";
			}
			iter = iter->parent;
		}
		*output_stream << "</td>";
	
		for (int i = 0; i < rule_constraints.size(); i++) {
			constraint = rule_constraints[i];
			*output_stream << "<td><b>" << constraint->abbreviation() << "=</b></td>";
			doubleInColor(constraint->value(rule), constraint->upperBound(), constraint->lowerBound(), constraint->zeroLimit(), 1, constraint->increasing());
		}
	
		*output_stream << "</tr>" << std::endl;
	}
}

/**
 * Outputs all association rules in frequent itemset tree starting from given root.
 * @param root The root of the itemset tree.
 */
void HtmlOutputGenerator::outputRules(FITree* root) {
	assert(root != NULL);

	outputRuleHeader();

	*output_stream << "<html>";
	*output_stream << "<body>";
	*output_stream << "<table border=\"1\">" << std::endl;

	queue<FITree*> itemsets;
	itemsets.push(root);

	FITree* iterator;
	vector<AssociationRule*>* rules;

	while(!itemsets.empty()) {
		iterator = itemsets.front();
		itemsets.pop();
		
		rules = iterator->associationRules();
		for (int i = 0 ; i < rules->size(); i++) {
			outputRule((*rules)[i]);
		}
		delete rules;
		
		for (int i = 0; i < iterator->children.size(); i++) {
			itemsets.push(iterator->children[i]);
		}
	}
	
	*output_stream << "</table>";
	*output_stream << "</body>";
	*output_stream << "</html>";
}

/**
 * Finalizes the generator.
 */
void HtmlOutputGenerator::finalize() {
/*	*header_stream << "<table border=\"1\">";
	*header_stream << "<tr><td>Transactions:</td><td>" << database->transactions() << "</td></tr>" << std::endl;
	*header_stream << "<tr><td>Frequent sets:</td><td>" << fsets->size() << "</td></tr>" << std::endl;
	*header_stream << "<tr><td><b>Constraint</b></td><td><b>Limit:</b></td></tr>" << std::endl;
	for (vector<SetConstraint*>::iterator cons_iter = set_constraints->begin(); cons_iter != set_constraints->end(); cons_iter++) { 
		if ((*cons_iter)->name() != "") {
			*header_stream << "<tr><b><td>" << (*cons_iter)->name() << ":</td><td>" << (*cons_iter)->constraintLimit() << "</td></b></tr>" << std::endl;
		}
	}
	
	for (vector<RuleConstraint*>::iterator cons_iter = rule_constraints->begin(); cons_iter != rule_constraints->end(); cons_iter++) { 
		if ((*cons_iter)->name() != "") {
			*header_stream << "<tr><b><td>" << (*cons_iter)->name() << ":</td><td>" << (*cons_iter)->constraintLimit() << "</td></b></tr>" << std::endl;
		}
	}
	*header_stream << "</table>";
	*header_stream << "</body>";
	*header_stream << "</html>";
	
	//output_stream->close();
	//header_stream->close();
	//frame_stream->close();
	
	delete output_stream;
	delete header_stream;
	delete frame_stream;*/
}
