/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <math.h>

#include "JMeasureConstraint.h"
#include "FITree.h"
#include "AssociationRule.h"

/**
 * Create a J-measure constraint with minimum J-measure of 0
 */
JMeasureConstraint::JMeasureConstraint() {
	last_jmeasure = 0;
	min_jmeasure = 0;
}

/**
 * Create a J-measure constraint with a minimum J-measure.
 * @param d The minimum J-measure.
 */
JMeasureConstraint::JMeasureConstraint(double d) {
	last_jmeasure = 0;
	min_jmeasure = d;
}

/**
 * Set the minimum J-measure of this constraint.
 * @param d The minimum J-measure.
 */
void JMeasureConstraint::setMinJMeasure(double d) {
	min_jmeasure = d;
}

/**
 * Returns the result of last evaluation.
 * @return The J-measure of the pattern evaluated last.
 */
double JMeasureConstraint::lastResult() {
	return last_jmeasure;
}

/**
 * Value an association rule.
 * @param arule The association rule.
 * @return The J-measure of the association rule.
 */
double JMeasureConstraint::value(AssociationRule* arule) {
	FITree* root = arule->itemset->root();
	if (arule->condition->support == root->support || arule->consequent->support == root->support) {
		last_jmeasure = 0;
	} else {
		last_jmeasure = (double)arule->itemset->support / (double)root->support	* log10((double)(arule->itemset->support * root->support) / (double)(arule->consequent->support * arule->condition->support)) + ((double)(arule->condition->support - arule->itemset->support) / (double)root->support) * log10(((double)(arule->condition->support - arule->itemset->support) / (double)root->support)/((double)(arule->condition->support * root->support - arule->condition->support * arule->consequent->support)/((double)root->support * root->support)));
	}
	return last_jmeasure;
}

/**
 * Validate an association rule.
 * @param arule The association rule.
 * @return True if the J-measure of the rule is higher than the minimum
 * J-measure of this constraint.
 */
bool JMeasureConstraint::validate(AssociationRule* arule) {
	if (value(arule) >= min_jmeasure) {
		return true;
	}
	
	return false;
}

/**
 * Returns the minimum J-measure.
 * @return The minimum J-measure.
 */
double JMeasureConstraint::constraintLimit() const {
	return min_jmeasure;
}

/**
 * Returns the name of this constraint.
 * @return The name.
 */
string JMeasureConstraint::name() const {
	return "J-measure";
}

/**
 * Returns the abbreviation of this constraint.
 * @return The abbreviation.
 */
string JMeasureConstraint::abbreviation() const {
	return "j";
}

/**
 * Returns the description of this constraint.
 * @return The description.
 */
string JMeasureConstraint::description() const {
	return "P(A,B)log(P(A,B)/P(A)P(B)) + (P(A) - P(A,B))log((P(A) - P(A,B))/(P(A) - P(A)P(B)))";
}
