/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
     
#include "LiftConstraint.h"
#include "FITree.h"
#include "AssociationRule.h"
#include "Item.h"

/**
 * Create a lift constraint with minimum lift of 0
 */
LiftConstraint::LiftConstraint() {
	last_lift = 0;
	min_lift = 0;
}

/**
 * Create a lift constraint with a minimum lift.
 * @param d The minimum lift.
 */
LiftConstraint::LiftConstraint(double d) {
	last_lift = 0;
	min_lift = d;
}

/**
 * Set the minimum lift of this constraint.
 * @param d The minimum lift.
 */
void LiftConstraint::setMinLift(double d) {
	min_lift = d;
}

/**
 * Returns the result of last evaluation.
 * @return The lift of the pattern evaluated last.
 */
double LiftConstraint::lastResult() {
	return last_lift;
}

/**
 * Value an itemset.
 * @param fset The itemset.
 * @return The lift of the itemset.
 */
double LiftConstraint::value(FITree* fset) {
	FITree* node = fset;
	FITree* root = fset->root();
	int transactions = root->support;

	FITree* temp;
	double expected_freq = 1;
	
//fset->root()->outputTree(std::cerr);
//std::cerr << "Itemset: " << *fset << std::endl;
	while (node != NULL) {
		for (int i = 0; i < node->items.size(); i++) {
//std::cerr << "Item: " << *((node->items)[i]) << std::endl;
			temp = root->findChild((node->items)[i]);
			// Might be null for closed itemsets.
			// Then you have to find the child with highest support that has the node in it.
			if (temp == NULL) {
				int temp_sup = 0;
				for (int j = 0; j < root->children.size(); j++) {
					if (root->children[j]->findItem((node->items)[i]) != NULL) {
						if (root->children[j]->support > temp_sup) {
							temp_sup = root->children[j]->support;
						}
					}
				}
				expected_freq = expected_freq * ((double)temp_sup / (double)transactions);
			} else {
//std::cerr << "Root->findChild(Item): " << *(temp) << " SUP: " << temp->support << std::endl;
				expected_freq = expected_freq * ((double)temp->support / (double)transactions);
			}
//std::cerr << "expected_freq" << expected_freq << std::endl;
		}
		node = node->parent;
	}
//std::cerr << std::endl;
	
	double itemset_freq = (double)fset->support / (double)fset->root()->support;
	
//std::cerr << "Itemset: " << *fset << std::endl;
//std::cerr << "itemset_freq / expected_freq: " << itemset_freq << " / " << expected_freq << std::endl;
	last_lift = itemset_freq / expected_freq;
	
	return last_lift;
}

/**
 * Value an association rule.
 * @param arule The association rule.
 * @return The lift of the association rule.
 */
double LiftConstraint::value(AssociationRule* arule) {
	last_lift = (double)arule->itemset->support / ((double)arule->itemset->root()->support * (double)arule->consequent->support / (double)arule->itemset->root()->support);
	return last_lift;
}

/**
 * Validate an itemset.
 * @param fset The itemset.
 * @return True if the lift of the itemset is higher than the minimum
 * lift of this constraint.
 */
bool LiftConstraint::validate(FITree* fset) {
	value(fset);
	
	if ((last_lift >= 1 && last_lift >= min_lift) || (last_lift < 1 && (1 / last_lift) >= min_lift)) {
		return true;
	}
	
	return false;
}

/**
 * Validate an association rule.
 * @param arule The association rule.
 * @return True if the lift of the rule is higher than the minimum
 * lift of this constraint.
 */
bool LiftConstraint::validate(AssociationRule* arule) {
	value(arule);
	
	if ((last_lift >= 1 && last_lift >= min_lift) || (last_lift < 1 && (1 / last_lift) >= min_lift)) {
		return true;
	}
	
	return false;
}

/**
 * Returns the minimum lift.
 * @return The minimum lift.
 */
double LiftConstraint::constraintLimit() const {
	return min_lift;
}

/**
 * Returns the name of this constraint.
 * @return The name.
 */
string LiftConstraint::name() const {
	return "Lift";
}

/**
 * Returns the abbreviation of this constraint.
 * @return The abbreviation.
 */
string LiftConstraint::abbreviation() const {
	return "l";
}

/**
 * Returns the description of this constraint.
 * @return The description.
 */
string LiftConstraint::description() const {
	return "Lift of a pattern is the relative frequency of the pattern divided by the expected frequency of the pattern.";
}

/**
 * Return the highest possible value of this constraint.
 * @return The "highest" value of lift, anything over this will be considered HIGH
 */
double LiftConstraint::upperBound() const {
	return 5;
}

/**
 * Return the lowest possible value of this constraint.
 * @return The "lowest" value of lift, anything under this will be considered LOW
 */
double LiftConstraint::lowerBound() const {
	return 0;
}

/**
 * Return the "zero point" of this constraint.
 * @return 1
 */
double LiftConstraint::zeroLimit() const {
	return 1;
}

