/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <iostream>

#include <queue>
using std::queue;

#include <vector>
using std::vector;

#include <algorithm>
using std::sort;

#include "NDI.h"
#include "AVTree.h"
#include "ContextTree.h"
#include "FITree.h"
#include "AssociationRule.h"
#include "Item.h"
#include "Cover.h"

/**
 * Create a new NDI
 */
NDI::NDI() {
}

/**
 * Destructor.
 */
NDI::~NDI() {
}

/**
 * Generates a set of non-derivable frequent itemsets from a attribute-value dataset.
 * @param dataset The dataset.
 * @param min_support The minimum support threshold
 */
FITree* NDI::generateNDI(AVTree* dataset, int min_support) {
	FITree* frequent_items = dataset->frequentItems(min_support);

	FITree* iterator;
	queue<FITree*> q;
	q.push(frequent_items);
	
	int q_size = q.size();
	int q_iterator = 0;

	FITree* candidate;
	
	while(!q.empty()) {
		iterator = q.front();
		q.pop();
		
		for (int i = 0; i < iterator->children.size(); i++) {
			for (int j = i + 1; j < iterator->children.size(); j++) {
				candidate = new FITree((iterator->children[j])->items[0], iterator->children[i]);

				// THE ONE USING THE NEW BOUNDING METHOD
				// FOR SOME REASON THE SUBSET CHECK PREVENTS THE BUG... GO FIGURE
				iterator->children[i]->addChild(candidate);
				if (candidate->subsetsFrequent()) {
					boundSupport(candidate);
					if (candidate->upper_bound >= min_support && candidate->upper_bound != candidate->lower_bound) {
						candidate->countSupport();
						if (candidate->support >= min_support) {
						} else {
							iterator->children[i]->removeChild();
							delete candidate;
						}
					} else {
						iterator->children[i]->removeChild();
						delete candidate;
					}
				} else {
					iterator->children[i]->removeChild();
					delete candidate;
				}

				// THE BUGGY ONE USING THE NEW BOUNDING METHOD :/
				/*iterator->children[i]->addChild(candidate);
				candidate->countSupport();

				if (candidate->support >= min_support) {
					boundSupport(candidate);

					if (candidate->upper_bound == candidate->lower_bound) {
						iterator->children[i]->removeChild();
						delete candidate;
					}
				} else {
					iterator->children[i]->removeChild();
					delete candidate;
				}*/
			}
			q.push(iterator->children[i]);
		}
	}
	return frequent_items;
}

/**
 * Generates a set of non-derivable frequent itemsets from context dataset.
 * @param dataset The dataset.
 * @param min_support The minimum support threshold
 */
FITree* NDI::generateNDI(ContextTree* dataset, int min_support) {
	FITree* frequent_items = dataset->frequentItems(min_support);

	FITree* iterator;
	queue<FITree*> q;
	q.push(frequent_items);
	
	int q_size = q.size();
	int q_iterator = 0;

	FITree* candidate;
	Cover* candidate_cover;
	
	while(!q.empty()) {
		iterator = q.front();
		q.pop();
		
		for (int i = 0; i < iterator->children.size(); i++) {
			for (int j = i + 1; j < iterator->children.size(); j++) {
				candidate = new FITree((iterator->children[j])->items[0], iterator->children[i]);

				// THE ONE USING THE NEW BOUNDING METHOD
				// FOR SOME REASON THE SUBSET CHECK PREVENTS THE BUG... GO FIGURE
				iterator->children[i]->addChild(candidate);
				if (candidate->subsetsFrequent()) {
					boundSupport(candidate);
					if (candidate->upper_bound >= min_support && candidate->upper_bound != candidate->lower_bound) {
						candidate_cover = candidate->getCover();
						candidate->support = dataset->contextSupport(candidate_cover);
						delete candidate_cover;
						if (candidate->support >= min_support) {
						} else {
							iterator->children[i]->removeChild();
							delete candidate;
						}
					} else {
						iterator->children[i]->removeChild();
						delete candidate;
					}
				} else {
					iterator->children[i]->removeChild();
					delete candidate;
				}
			}
			q.push(iterator->children[i]);
		}
	}
	return frequent_items;
}

/**
 * Binds the support (finds the upper and lower bound of support) for an itemset
 * given it's subsets.
 * @param itemset The itemset
 * @return the width of the interval
 */
int NDI::boundSupport(FITree* itemset) {
	vector<FITree*>* subset = itemset->subsets();
	int* deduction_rule = new int[subset->size()];
	int number_of_subsets = subset->size();
	
	// Initialize rules
	deduction_rule[0] = 0;
	for (int i = 1; i < number_of_subsets; i++) {
		if ((itemset->height - (*subset)[i]->height) % 2) {
			// If the difference of size and subset size is odd then +
			deduction_rule[i] = (*subset)[i]->support;
		} else {
			// If it is even then -
			deduction_rule[i] = -(*subset)[i]->support;
		}
	}

	// Compute bounds
	// IMPROVEMENT: this probably could be improved by checking the rules i / 2 -- i  first rules
	// during the computation and then stopping the computation if the upper bound equals the lower bound
	// similar improvements could be done for bounding the confidence
	int middle;
	for (int i = 2; i <= number_of_subsets; i = i * 2) {
		for (int j = 0; j < number_of_subsets; j = j + i) {
			middle = (j + (j + i)) / 2;
			for (int k = middle;  k < j + i; k++) {
				deduction_rule[k] = deduction_rule[k] + deduction_rule[j + (k - middle)];
			}
		}
	}
	
	// Set the bounds
	for (int i = 0; i < number_of_subsets; i++) {
		if ((itemset->height - (*subset)[i]->height) % 2) {
			// Not an equal division by 2 means an upper bound
			//std::cerr << "upperbound for itemset " << *itemset << ":" << temp_bound << std::endl;
			if (deduction_rule[i] < itemset->upper_bound) {
				itemset->upper_bound = deduction_rule[i];
			}
		} else {
			// An equal division by 2 means a lower bound
			//std::cerr << "lowerbound for itemset " << *itemset << ":" << temp_bound << std::endl;
			if (deduction_rule[i] > itemset->lower_bound) {
				itemset->lower_bound = deduction_rule[i];
			}
		}
	}
	delete deduction_rule;

#ifndef _STORE_SUBSETS
	delete subset;
#endif	
	return (itemset->upper_bound - itemset->lower_bound);
}

/**
 * Binds the support of the condition of an association rule
 * using the supports of its _available_ condition subrules given an association rule and the subsets of
 * condition of that rule.
 * @param arule The association rule.
 * @param condition_subsets The condition subsets
 * @return the width of the support interval
 */
int NDI::boundConditionSupport(AssociationRule* arule, vector<FITree*>* condition_subsets) {
	FITree* itemset = arule->condition;
	itemset->upper_bound = INT_MAX;
	itemset->lower_bound = 0;
	vector<FITree*>* subset = condition_subsets;
	int number_of_subsets = subset->size();

	int* deduction_rule = new int[subset->size()];
	bool* valid_rule = new bool[subset->size()];
	
	// Initialize rules
	deduction_rule[0] = 0;
	valid_rule[0] = true;
	for (int i = 1; i < number_of_subsets; i++) {
		// If there is a non-pruned condition subrule with subset i
		// then using the subset support is valid.
		if ((*subset)[i]->consequent_rules[arule->consequent] == NULL) {
			valid_rule[i] = false;
		} else {
			valid_rule[i] = true;
		}	
		
		// Initialize supports of subsets
		if ((itemset->height - (*subset)[i]->height) % 2) {
			// If the difference of size and subset size is odd then +
			deduction_rule[i] = (*subset)[i]->support;
		} else {
			// If it is even then -
			deduction_rule[i] = -(*subset)[i]->support;
		}
	}

	// Compute bounds
	int middle;
	for (int i = 2; i <= number_of_subsets; i = i * 2) {
		for (int j = 0; j < number_of_subsets; j = j + i) {
			middle = (j + (j + i)) / 2;
			for (int k = middle;  k < j + i; k++) {
				// If we are adding an invalid support then all subsequent deduction rules are also invalid
				if (!valid_rule[j + (k - middle)]) {
					valid_rule[k] = false;
				} else {
					deduction_rule[k] = deduction_rule[k] + deduction_rule[j + (k - middle)];
				}
			}
		}
	}
	
	// Set the bounds
	for (int i = 0; i < number_of_subsets; i++) {
		if (valid_rule[i]) {
			if ((itemset->height - (*subset)[i]->height) % 2) {
				// Not an equal division by 2 means an upper bound
				//std::cerr << "upperbound for itemset " << *itemset << ":" << temp_bound << std::endl;
				if (deduction_rule[i] < itemset->upper_bound) {
					itemset->upper_bound = deduction_rule[i];
				}
			} else {
				// An equal division by 2 means a lower bound
				//std::cerr << "lowerbound for itemset " << *itemset << ":" << temp_bound << std::endl;
				if (deduction_rule[i] > itemset->lower_bound) {
					itemset->lower_bound = deduction_rule[i];
				}
			}
		}
	}
	
	// The condition can not have zero support, since the rule is frequent.
	if (itemset->lower_bound == 0) {
		itemset->lower_bound = 1;
	}
	
	// The condition has to be less than the number of transactions in dataset
	if (itemset->upper_bound == INT_MAX) {
		FITree* node = itemset;
		while (node->parent != NULL) {
			node = node->parent;
		}
		itemset->upper_bound = node->support;
	}

	delete valid_rule;
	delete deduction_rule;
	
	return (itemset->upper_bound - itemset->lower_bound);
}

/**
 * Binds the support of the association rule using the _available_ condition subrules.
 * @param arule The association rule
 * @return The confidence interval.
 */
Interval* NDI::boundConfidenceCondition(AssociationRule* arule) {
	if (arule->condition->height <= 1) {
		// If the condition is a singleton or an empty set, we can not say anything
		// about the confidence of the rule using the condition subrules
		return new Interval(0,1);
	}
	
	vector<FITree*>* condition_subsets = arule->condition->subsets();
	int number_of_subsets = condition_subsets->size() * 2;
	int* deduction_rule = new int[number_of_subsets];
	int* size_difference = new int[number_of_subsets];
	bool* valid_rule = new bool[number_of_subsets];
	bool* condition_rule = new bool[number_of_subsets];
	
	AssociationRule* r;
	
	// Initialize rules for association rule "ab => cd"
	// Initialize deduction_rule("{a,b,c,d}")
	deduction_rule[0] = 0;
	size_difference[0] = 0;
	valid_rule[0] = true;
	condition_rule[0] = false;

	// Initialize deduction_rule("condition_subset Union {c,d}")
	for (int i = 1; i < number_of_subsets / 2; i++) {
		// support is obtained from a rule with same condition as the subset	
		r = arule->consequent->condition_rules[(*condition_subsets)[i]];
		if (r != NULL) {
			size_difference[i] = (arule->condition->height + 1) - (r->condition->height + 1);
			if (size_difference[i] % 2) {
				// If the difference of size and subset size is odd then +
				deduction_rule[i] = r->itemset->support;
			} else {
				// If the difference of size and subset size is odd then -
				deduction_rule[i] = -r->itemset->support;
			}
			valid_rule[i] = true;
			condition_rule[i] = false;
		} else {
			valid_rule[i] = false;
		}
	}

	// Initialize deduction_rule("condition")
	boundConditionSupport(arule, condition_subsets);
	deduction_rule[number_of_subsets / 2] = arule->condition->lower_bound;
	size_difference[number_of_subsets / 2] = 1;
	valid_rule[number_of_subsets / 2] = true;
	condition_rule[number_of_subsets / 2] = true;

	// Initialize deduction_rule("condition_subset")
	for (int i = number_of_subsets / 2 + 1; i < number_of_subsets; i++) {
		// support is obtained from a rule with same condition as the subset	
		r = arule->consequent->condition_rules[(*condition_subsets)[i - number_of_subsets / 2]];
		if (r != NULL) {
			size_difference[i] = (arule->condition->height + 1) - (r->condition->height);
			if (size_difference[i] % 2) {
				// If the difference of size and subset size is odd then +
				deduction_rule[i] = r->condition->support;
			} else {
				// If the difference of size and subset size is odd then -
				deduction_rule[i] = -r->condition->support;
			}
			valid_rule[i] = true;
			condition_rule[i] = false;
		} else {
			valid_rule[i] = false;
		}
	}
	// End initialization
	
	// Compute bounds
	int middle;
	for (int i = 2; i <= number_of_subsets; i = i * 2) {
		for (int j = 0; j < number_of_subsets; j = j + i) {
			middle = (j + (j + i)) / 2;
			for (int k = middle;  k < j + i; k++) {
				// If we are adding the support of the condition, then all subsequent deduction rules contain that support too
				if (condition_rule[j + (k - middle)]) {
					condition_rule[k] = true;
				}
				// If we are adding an invalid support then all subsequent deduction rules are also invalid
				if (!valid_rule[j + (k - middle)]) {
					valid_rule[k] = false;
				} else {
					deduction_rule[k] = deduction_rule[k] + deduction_rule[j + (k - middle)];
				}
			}
		}
	}
	
	// Set the bounds for each possible support of condition
	double temp_conf;
	int rule_supp_lower;
	int rule_supp_upper;
	Interval* confidence_interval = new Interval(2,-1);

	for (int cond_supp = arule->condition->lower_bound; cond_supp <= arule->condition->upper_bound; cond_supp++) {
		rule_supp_lower = 0;
		rule_supp_upper = INT_MAX;
		// Search the lower and upper bound of support of the itemset for condition support = cond_supp
		// IMPROVEMENT: this probably could be improved by storing the indexes of valid deduction rules and
		// using them.
		for (int i = 0; i < number_of_subsets; i++) {
			if (valid_rule[i]) {
				if (size_difference[i] % 2) {
					// Not an equal division by 2 means an upper bound
					if (deduction_rule[i] < rule_supp_upper) {
						rule_supp_upper = deduction_rule[i];
					}
				} else {
					// An equal division by 2 means a lower bound
					if (deduction_rule[i] > rule_supp_lower) {
						rule_supp_lower = deduction_rule[i];
					}
				}
				// Next round the valid condition rules have 1 higher support
				// this corresponds to iterating all possible condition supports
				if (condition_rule[i]) {
					deduction_rule[i]++;
				}
			}
		}

		// It is possible that the lower bound of rule support is higher than the upper bound of rule support
		// in some occasions, then the computed bounds are not valid...
		// example dataset:
		// gender age disease
		// male   0   affected
		// male   50  control
		// male   50  control
		// female 0   affected
		// female 100 affected
		// Now rule {gender=male, disease=control} => {age=50}
		// has for supp({gender=male, disease=control}) bounds [1,3]
		// and for supp({gender=male, disease=control}) = 1 the bounds for
		// supp({gender=male, disease=control, age=50}) the bounds are [2,1]!

		if (rule_supp_upper >= rule_supp_lower) {
			temp_conf = (double)rule_supp_upper/(double)cond_supp;
			// Take the maximum upper bound and minimum lower bound of all possible confidence bounds
			if (confidence_interval->upper_bound < temp_conf) {
				confidence_interval->upper_bound = temp_conf;
			}
		
			temp_conf = (double)rule_supp_lower/(double)cond_supp;
			if (confidence_interval->lower_bound > temp_conf) {
				confidence_interval->lower_bound = temp_conf;
			}
		}

		/*
		// SEE FOR YOURSELF THAT THIS IS POSSIBLE!
		if (rule_supp_upper < rule_supp_lower) {
			std::cerr << "\n\ncond_supp:" << cond_supp << std::endl;
			std::cerr << "rule_supp_upper:" << rule_supp_upper << std::endl;
			std::cerr << "rule_supp_lower:" << rule_supp_lower << std::endl;
			std::cerr << *arule << std::endl;
			std::cerr << *(arule->condition) << "[" <<  arule->condition->lower_bound << "," << arule->condition->upper_bound << "]" << std::endl;
			for (int i = 0; i < number_of_subsets; i++) {
				if (size_difference[i] % 2) {
					// Not an equal division by 2 means an upper bound
					if (i < number_of_subsets / 2) {
						std::cerr << *(arule->consequent) << " " << *((*condition_subsets)[i]) << std::endl;
					} else {
						std::cerr << *((*condition_subsets)[i - number_of_subsets / 2]) << std::endl;
					}
					if (condition_rule[i]) {
						std::cerr << "valid= " << valid_rule[i] << " <= deduction_rule[" << i << "] = " << deduction_rule[i] - 1 << std::endl;
					} else {
						std::cerr << "valid= " << valid_rule[i] << " <= deduction_rule[" << i << "] = " << deduction_rule[i] << std::endl;
					}
				} else {
					// An equal division by 2 means a lower bound
					if (i < number_of_subsets / 2) {
						std::cerr << *(arule->consequent) << " " << *((*condition_subsets)[i]) << std::endl;
					} else {
						std::cerr << *((*condition_subsets)[i - number_of_subsets / 2]) << std::endl;
					}
					if (condition_rule[i]) {
						std::cerr << "valid= " << valid_rule[i] << " >= deduction_rule[" << i << "] = " << deduction_rule[i] - 1 << std::endl;
					} else {
						std::cerr << "valid= " << valid_rule[i] << " >= deduction_rule[" << i << "] = " << deduction_rule[i] << std::endl;
					}
				}
			}
		}
		*/
	}
	
#ifndef _STORE_SUBSETS
	delete condition_subsets;
#endif
	delete deduction_rule;
	delete size_difference;
	delete valid_rule;
	delete condition_rule;

	// If this is two (as set earlier), then there are no valid rules to lower bound the confidence
	if (confidence_interval->lower_bound == 2) {
		confidence_interval->lower_bound = 0;
	}
	
	// If this is minus 1 (as set earlier), then there are no valid rules to upper bound the confidence
	if (confidence_interval->upper_bound == -1) {
		confidence_interval->upper_bound = 1;
	}
	
	return confidence_interval;
}

/**
 * Binds the support of the association rule using the _available_ consequent subrules.
 * @param arule The association rule
 * @return The confidence interval.
 */
Interval* NDI::boundConfidenceConsequent(AssociationRule* arule) {
	if (arule->consequent->height <= 1) {
		// If the consequent is a singleton or an empty set, we can not say anything
		// about the confidence of the rule using the consequent subrules
		return new Interval(0,1);
	}

	vector<FITree*>* consequent_subsets = arule->consequent->subsets();
	int number_of_subsets = consequent_subsets->size();
	int* deduction_rule = new int[number_of_subsets];
	bool* valid_rule = new bool[number_of_subsets];
	
	AssociationRule* r;
	
	// Initialize rules for association rule "ab => cd"
	// Initialize deduction_rule("{a,b,c,d}")
	deduction_rule[0] = 0;
	valid_rule[0] = true;

	// Initialize deduction_rule("{a,b} Union consequent_subset")
	for (int i = 1; i < number_of_subsets; i++) {
		r = arule->condition->consequent_rules[(*consequent_subsets)[i]];
		if (r != NULL) {
			if ((arule->consequent->height - (*consequent_subsets)[i]->height) % 2) {
				// If the difference of size and subset size is odd then +
				deduction_rule[i] = r->itemset->support;
			} else {
				// If the difference of size and subset size is odd then -
				deduction_rule[i] = -r->itemset->support;
			}
			valid_rule[i] = true;
		} else {
			valid_rule[i] = false;
		}
	}
	
	// Compute bounds
	int middle;
	for (int i = 2; i <= number_of_subsets; i = i * 2) {
		for (int j = 0; j < number_of_subsets; j = j + i) {
			middle = (j + (j + i)) / 2;
			for (int k = middle;  k < j + i; k++) {
				// If we are adding an invalid support then all subsequent deduction rules are also invalid
				if (!valid_rule[j + (k - middle)]) {
					valid_rule[k] = false;
				} else {
					deduction_rule[k] = deduction_rule[k] + deduction_rule[j + (k - middle)];
				}
			}
		}
	}
	
	Interval* confidence_interval = new Interval(0,1);
	double temp_conf;
	
	// Search the lower and upper bound of confidence
	for (int i = 0; i < number_of_subsets; i++) {
		if (valid_rule[i]) {
			temp_conf = (double)deduction_rule[i] / (double)arule->condition->support;
			if ((arule->consequent->height - (*consequent_subsets)[i]->height) % 2) {
				// Not an equal division by 2 means an upper bound
				if (temp_conf < confidence_interval->upper_bound) {
					confidence_interval->upper_bound = temp_conf;
				}
			} else {
				// An equal division by 2 means a lower bound
				if (temp_conf > confidence_interval->lower_bound) {
					confidence_interval->lower_bound = temp_conf;
				}
			}
		}
	}

#ifndef _STORE_SUBSETS
	delete consequent_subsets;
#endif
	delete deduction_rule;
	delete valid_rule;

	return confidence_interval;
}

/**
 * Binds the confidence of an association rule given its available subrules.
 * @param arule The association rule
 * @return The confidence interval.
 */ 
Interval* NDI::boundConfidence(AssociationRule* arule) {
	Interval* consequent_interval = boundConfidenceConsequent(arule);
	
	if (consequent_interval->upper_bound == consequent_interval->lower_bound) {
		return consequent_interval;
	}
	
	Interval* condition_interval = boundConfidenceCondition(arule);

	if (consequent_interval->lower_bound > condition_interval->lower_bound) {
		condition_interval->lower_bound = consequent_interval->lower_bound;
	} 

	if (consequent_interval->upper_bound < condition_interval->upper_bound) {
		condition_interval->upper_bound = consequent_interval->upper_bound;
	} 
	
	delete consequent_interval;
	return condition_interval;
}

/**
 * Creates a new interval with lower bound of 0 and upper bound of 1.
 */
Interval::Interval() {
	lower_bound = 0;
	upper_bound = 1;
}

/**
 * Creates a new interval with given lower and upper bound.
 * @param l The lower bound
 * @param u The upper bound
 */
Interval::Interval(double l, double u) {
	lower_bound = l;
	upper_bound = u;
}

/**
 * Destructor.
 */
Interval::~Interval() {
}

