/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "PhpOutputGenerator.h"
#include "SetConstraint.h"
#include "RuleConstraint.h"

#include "Item.h"
#include "FITree.h"
#include "AssociationRule.h"

#include <iomanip>
#include <iostream>
using std::ostream;
using std::endl;
using std::cout;

#include <queue>
using std::queue;

/**
 * Creates a new html output generator.
 */
PhpOutputGenerator::PhpOutputGenerator() {
	output_name = "";
	output_stream = &cout;
	output_stream->setf(std::ios::fixed);
	output_stream->precision(3);
}

/**
 * Set the output name.
 * @param n The name.
 */
void PhpOutputGenerator::setOutputName(string n) {
	assert(n != "");

	output_name = n;
	
	if (output_stream != NULL && output_stream != &cout) {
		delete output_stream;
	}	

	output_stream = new ofstream(output_name.c_str());	
	output_stream->setf(std::ios::fixed);
}


/**
 * Prints a double in color (html) using default values.
 * @param d The double
 */
void PhpOutputGenerator::doubleInColor(double d) {
	doubleInColor(d, 1.0, -1.0, 0.0, 1, true);
}

/**
 * Outputs a double in color (html) using given upper and lower bound, zero limit, spacing and increasing
 * @param d The double to be printed.
 * @param upper_b The upper bound of the double (closer to the upper bound the double is, the greener the color)
 * @param upper_b The lower bound of the double (closer to the lower bound the double is, more red the color)
 * @param negative_l The limit where the color is white.
 * @param spacing
 * @param increasing Tells whether low or high values should be green.
 */
void PhpOutputGenerator::doubleInColor(double d, double upper_b, double lower_b, double negative_l, int spacing, bool increasing) {
	double upper_bound;
	double lower_bound;
	double negative_limit;
	double value;
	
	if (increasing) {
		upper_bound = upper_b;
		lower_bound = lower_b;
		negative_limit = negative_l;
		value = d;
	} else {
		upper_bound = -lower_b;
		lower_bound = -upper_b;
		negative_limit = -negative_l;
		value = -d;
	}

	//char buffer[10];
	//gcvt(d, 3, buffer);
	
	if (value > upper_bound) {
		value = upper_bound;
	}
	
	//std::cerr << "double:" << value << " upper_bound:" << upper_bound << " lower_bound:" << lower_bound << " negative_limit:" << negative_limit;
	if (d >= negative_limit) {
		int int_result = static_cast<int>(((value - negative_limit) / (upper_bound - negative_limit)) * 255 / spacing) * spacing;;
		//std::cerr << " int_result: " << int_result << std::endl;
		*output_stream << "<td class=\"measure\" bgcolor=\"#";
		*output_stream << "0123456789ABCDEF"[static_cast<int>((255 - int_result) / 16)];
		*output_stream << "0123456789ABCDEF"[(255 - int_result) % 16];
		*output_stream << "FF";
		*output_stream << "0123456789ABCDEF"[static_cast<int>((255 - int_result) / 16)];
		*output_stream << "0123456789ABCDEF"[(255 - int_result) % 16] << "\">" << d << "</td>";
	} else {
		int int_result = static_cast<int>(((value - negative_limit) / (negative_limit - lower_bound)) * 255 / spacing) * spacing;;
		//cerr << "int_result:" << int_result << "\n";
		//std::cerr << " int_result: " << int_result << std::endl;
		*output_stream << "<td class=\"measure\" bgcolor=\"#";
		*output_stream << "FF";
		*output_stream << "0123456789ABCDEF"[static_cast<int>((255 + int_result) / 16)];
		*output_stream << "0123456789ABCDEF"[(255 + int_result) % 16];
		*output_stream << "0123456789ABCDEF"[static_cast<int>((255 + int_result) / 16)];
		*output_stream << "0123456789ABCDEF"[(255 + int_result) % 16] << "\">" << d << "</td>";
	}
}


/**
 * Outputs the frame.
 */
void PhpOutputGenerator::outputHeader() {		
	*output_stream << "<head>" << endl;
	*output_stream << "	<style type=\"text/css\">" << endl;
	*output_stream << "	div.pos_top {" << endl;
	*output_stream << "		height: 20%;" << endl;
	*output_stream << "		width: 100%;" << endl;
	*output_stream << "	}" << endl;
	*output_stream << "	div.pos_bottom {" << endl;
	*output_stream << "		overflow: scroll;" << endl;
	*output_stream << "		height: 80%;" << endl;
	*output_stream << "		width: 100%;" << endl;
	*output_stream << "	}" << endl;
	*output_stream << "	</style>" << endl;
	*output_stream << "</head>" << endl;
}

/**
 * Outputs the php sorting script
 */
void PhpOutputGenerator::outputPhpSort() {
	*output_stream << "<?php" << endl;
	*output_stream << "	} else {" << endl;
	*output_stream << "		$handle = fopen($_GET['rule_file'], \"r\");" << endl;
	*output_stream << "		$contents = fread($handle, filesize($_GET['rule_file']));" << endl;
	*output_stream << "		fclose($handle);" << endl;
	*output_stream << endl;
	*output_stream << "		$reg_exp_pattern = '/^<tr class=\"pattern\">/';" << endl;
	*output_stream << "		$reg_exp_measure = '/<td class=\"measure\".*?>(.*?)<\\/td>/';" << endl;
	*output_stream << endl;
	*output_stream << "		$lines = explode(\"\\n\", $contents);" << endl;;
	*output_stream << "		$line_array;" << endl;
	*output_stream << "		foreach ($lines as $line) {" << endl;
	*output_stream << "			if (preg_match($reg_exp_pattern, $line)) {" << endl;
	*output_stream << "				preg_match_all($reg_exp_measure, $line, $matches, PREG_SET_ORDER);" << endl;
	*output_stream << "				$line_array[$line] = $matches[$_GET['column']][1];" << endl;
	*output_stream << "			}" << endl;
	*output_stream << "		}" << endl;
	*output_stream << endl;
	*output_stream << "		if ($_GET['descending'] == 1) {" << endl;
	*output_stream << "			if ($_GET['sort_type'] == \"lexical\") {" << endl;
	*output_stream << "				arsort($line_array, SORT_STRING);" << endl;
	*output_stream << "			} else {" << endl;
	*output_stream << "				arsort($line_array, SORT_NUMERIC);" << endl;
	*output_stream << "			}" << endl;
	*output_stream << "		} else {" << endl;
	*output_stream << "			if ($_GET['sort_type'] == \"lexical\") {" << endl;
	*output_stream << "				asort($line_array, SORT_STRING);" << endl;
	*output_stream << "			} else {" << endl;
	*output_stream << "				asort($line_array, SORT_NUMERIC);" << endl;
	*output_stream << "			}" << endl;
	*output_stream << "		}" << endl;
	*output_stream << endl;
	*output_stream << "		foreach ($line_array as $line => $measure) {" << endl;
	*output_stream << "			echo $line;" << endl;
	*output_stream << "		}" << endl;
	*output_stream << "	}" << endl;
	*output_stream << "?>" << endl;
}

/**
 * Outputs the header for association rule output.
 */
void PhpOutputGenerator::outputRuleMeasures() {
	*output_stream << "	<div class=\"pos_top\">" << endl;
	*output_stream << "	<table border=\"1\">" << endl;
	*output_stream << "	<tr>";
	*output_stream << "<td><a href=\"" << output_name << "?rule_file=" << output_name << "&sort_type=lexical&column=0&descending=1\"><b>" << "Condition" << "</b></a></td>";
	*output_stream << "<td><b>" << "->" << "</b></td>";
	*output_stream << "<td><a href=\"" << output_name << "?rule_file=" << output_name << "&sort_type=lexical&column=1&descending=1\"><b>" << "Consequent" << "</b></a></td>";
	int column_number = 2;
		
	for (vector<RuleConstraint*>::iterator cons_iter = rule_constraints.begin(); cons_iter != rule_constraints.end(); cons_iter++) { 
		if ((*cons_iter)->name() != "") {
			*output_stream << "<td><a href=\"" << output_name << "?rule_file=" << output_name << "&sort_type=numeric" << "&column=" << column_number << "&descending=" << (*cons_iter)->increasing() << "\"><b>" << (*cons_iter)->name() << " (" << (*cons_iter)->abbreviation() << ")" << "</b></a></td>";
			column_number++;
		}
	}
	*output_stream << "</tr>" << endl;

	*output_stream << "	<tr>";
	*output_stream << "<td><b>" << "" << "</b></td>";
	*output_stream << "<td><b>" << "" << "</b></td>";
	*output_stream << "<td><b>" << "" << "</b></td>";

	for (vector<RuleConstraint*>::iterator cons_iter = rule_constraints.begin(); cons_iter != rule_constraints.end(); cons_iter++) { 
		if ((*cons_iter)->name() != "") {
			*output_stream << "<td><b>" << (*cons_iter)->description() << "</b></td>";
		}
	}
	*output_stream << endl << "	</tr></table></div>" << endl;
}

/**
 * Outputs the header for itemset output.
 */
void PhpOutputGenerator::outputItemsetMeasures() {
	*output_stream << "	<div class=\"pos_top\">" << endl;
	*output_stream << "	<table border=\"1\">" << endl;
	*output_stream << "	<tr>";
	*output_stream << "<td><a href=\"" << output_name << "?rule_file=" << output_name << "&sort_type=lexical&column=0&descending=1\"><b>" << "Itemset" << "</b></a></td>";
	int column_number = 1;
		
	for (int i = 0; i < set_constraints.size(); i++) {
		*output_stream << "<td><a href=\"" << output_name << "?rule_file=" << output_name << "&sort_type=numeric" << "&column=" << column_number << "&descending=" << set_constraints[i]->increasing() << "\"><b>" << set_constraints[i]->name() << " (" << set_constraints[i]->abbreviation() << ")" << "</b></a></td>";
		column_number++;
	}

	*output_stream << "</tr>" << endl;

	*output_stream << "	<tr>";
	*output_stream << "<td><b></b></td>";

	for (int i = 0; i < set_constraints.size(); i++) {
		*output_stream << "<td><b>" << set_constraints[i]->description() << "</b></td>";
	}

	*output_stream << endl << "	</tr></table></div>" << endl;
}

/**
 * Add a set constraint to this output generator.
 * @param c The set constraint.
 */
void PhpOutputGenerator::addSetConstraint(SetConstraint* c) {
	assert(c != NULL);
	set_constraints.push_back(c);
}

/**
 * Add a rule constraint to this output generator.
 * @param c The rule constraint.
 */
void PhpOutputGenerator::addRuleConstraint(RuleConstraint* c) {
	assert(c != NULL);
	rule_constraints.push_back(c);
}

/**
 * Removes a constraint by it's name.
 * @param s The constraint name.
 */
void PhpOutputGenerator::removeConstraint(string s) {
	for (int i = 0; i < set_constraints.size(); i++) {
		if (set_constraints[i]->name() == s) {
			set_constraints.erase(set_constraints.begin() + i);
		}
	}

	for (int i = 0; i < rule_constraints.size(); i++) {
		if (rule_constraints[i]->name() == s) {
			rule_constraints.erase(rule_constraints.begin() + i);
		}
	}
}

/**
 * Outputs an itemset.
 * @param set The itemset.
 */
void PhpOutputGenerator::outputSet(FITree* set) {
	assert(set != NULL);

	SetConstraint* constraint;
	
	bool valid = true;
	
	for (int i = 0; i < set_constraints.size() && valid; i++) {
		constraint = set_constraints[i];
		if (!constraint->validate(set)) {
			valid = false;
		}
	}

	if (valid) {

		*output_stream << "<tr class=\"pattern\">";
		*output_stream << "<td class=\"measure\">";
	
		FITree* iter = set;
		while (iter != NULL) {
			for (int i = 0; i < iter->items.size(); i++) {
				*output_stream << *(iter->items[i]) << "<br>";
			}
			iter = iter->parent;
		}
	
		*output_stream << "</td>";
	
		for (int i = 0; i < set_constraints.size(); i++) {
			constraint = set_constraints[i];
			*output_stream << "<td><b>" << constraint->abbreviation() << "=</b></td>";
			output_stream->precision(constraint->precision());
			doubleInColor(constraint->value(set), constraint->upperBound(), constraint->lowerBound(), constraint->zeroLimit(), 1, constraint->increasing());
		}
	
		*output_stream << "</tr>" << std::endl;
	}
	
	for (int i = 0; i < set->children.size(); i++) {
		outputSet(set->children[i]);
	}
}

/**
 * Outputs all itemsets strating from given frequent itemset root.
 * @param set The root of the tree.
 */
void PhpOutputGenerator::outputSets(FITree* set) {
	assert(set != NULL && output_stream != NULL);

	*output_stream << "#!/usr/local/bin/php" << endl;
	*output_stream << "<html>" << endl;
	outputHeader();

	*output_stream << "<body>" << endl;
	outputItemsetMeasures();
	
	*output_stream << "	<div class=\"pos_bottom\">" << endl;
	*output_stream << "	<table border=\"1\">" << endl;
	*output_stream << "<?php" << endl;
	*output_stream << "	if (!(isset($_GET['rule_file']) && isset($_GET['sort_type']) && isset($_GET['column']) && isset($_GET['descending']))) {" << endl;
	*output_stream << "?>" << endl;

	outputSet(set);
	
	outputPhpSort();
	
	*output_stream << "	</table>" << endl;
	*output_stream << "	</div>" << endl;
	*output_stream << "</body>" << endl;
	*output_stream << "</html>" << endl;
}

/**
 * Outputs an association rule.
 * @param rule The association rule.
 */
void PhpOutputGenerator::outputRule(AssociationRule* rule) {
	assert(rule != NULL);

	RuleConstraint* constraint;
	
	bool valid = true;
	
	for (int i = 0; i < rule_constraints.size() && valid; i++) {
		constraint = rule_constraints[i];
		if (!constraint->validate(rule)) {
			valid = false;
		}
	}

	if (valid) {

		*output_stream << "<tr class=\"pattern\">";
		*output_stream << "<td class=\"measure\">";
	
		FITree* iter = rule->condition;
		while (iter != NULL) {
			for (int i = 0; i < iter->items.size(); i++) {
				*output_stream << *(iter->items[i]) << "<br>";
			}
			iter = iter->parent;
		}
	
		*output_stream << "</td><td>-></td>";
		*output_stream << "<td class=\"measure\">";
	
		iter = rule->consequent;
		while (iter != NULL) {
			for (int i = 0; i < iter->items.size(); i++) {
				*output_stream << *(iter->items[i]) << "<br>";
			}
			iter = iter->parent;
		}
		*output_stream << "</td>";
	
		for (int i = 0; i < rule_constraints.size(); i++) {
			constraint = rule_constraints[i];
			*output_stream << "<td><b>" << constraint->abbreviation() << "=</b></td>";
			output_stream->precision(constraint->precision());
			doubleInColor(constraint->value(rule), constraint->upperBound(), constraint->lowerBound(), constraint->zeroLimit(), 1, constraint->increasing());
		}
	
		*output_stream << "</tr>" << std::endl;
	}
}

/**
 * Outputs all association rules in frequent itemset tree starting from given root.
 * @param root The root of the itemset tree.
 */
void PhpOutputGenerator::outputRules(FITree* root) {
	assert(root != NULL);

	*output_stream << "#!/usr/local/bin/php" << endl;
	*output_stream << "<html>" << endl;
	outputHeader();

	*output_stream << "<body>" << endl;
	outputRuleMeasures();

	*output_stream << "	<div class=\"pos_bottom\">" << endl;
	*output_stream << "	<table border=\"1\">" << endl;
	*output_stream << "<?php" << endl;
	*output_stream << "	if (!(isset($_GET['rule_file']) && isset($_GET['sort_type']) && isset($_GET['column']) && isset($_GET['descending']))) {" << endl;
	*output_stream << "?>" << endl;

	queue<FITree*> itemsets;
	itemsets.push(root);

	FITree* iterator;
	vector<AssociationRule*>* rules;

	while(!itemsets.empty()) {
		iterator = itemsets.front();
		itemsets.pop();
		
		rules = iterator->associationRules();
		for (int i = 0 ; i < rules->size(); i++) {
			outputRule((*rules)[i]);
		}
		delete rules;
		
		for (int i = 0; i < iterator->children.size(); i++) {
			itemsets.push(iterator->children[i]);
		}
	}
	
	outputPhpSort();
	
	*output_stream << "	</table>" << endl;
	*output_stream << "	</div>" << endl;
	*output_stream << "</body>" << endl;
	*output_stream << "</html>" << endl;
}

/**
 * Finalizes the generator.
 */
void PhpOutputGenerator::finalize() {
}
