/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <queue>
using std::queue;

#include "RuleGenerator.h"
#include "AssociationRule.h"
#include "FITree.h"
#include "Item.h"
#include "NDI.h"

/**
 * Creates a new rule generator.
 */
RuleGenerator::RuleGenerator() {
	rules_generated = 0;
	rules_output = 0;
	rules_with_zero_dob = 0;
	rules_with_high_confidence = 0;
	min_difference_of_bounds = -1;
	min_confidence = 0;
}

/**
 * The destructor.
 */
RuleGenerator::~RuleGenerator() {
}

/**
 * Generates and counts non-derivable rules from a frequent itemset tree.
 * @param root The root of the frequent itemset tree.
 * @return The number of rules generated.
 */	
int RuleGenerator::generateRules(FITree* root) {
	queue<FITree*> itemsets;
	itemsets.push(root);

	FITree* iterator;
	vector<AssociationRule*>* rules;
	vector<AssociationRule*>* subrules;
	NDI ndi;

	while(!itemsets.empty()) {
		iterator = itemsets.front();
		itemsets.pop();
		
		rules = iterator->associationRules();
		for (int i = 0 ; i < rules->size(); i++) {
			// Only Non-Empty consequents
			if ((*rules)[i]->consequent->parent !=NULL) {
				double confidence = (double)((double)(*rules)[i]->itemset->support / (double)(*rules)[i]->condition->support);
				if (confidence >= min_confidence) {
					rules_with_high_confidence++;
				
					//Interval* interval = ndi.boundConfidenceCondition((*rules)[i]);
					//Interval* interval = ndi.boundConfidenceConsequent((*rules)[i]);
					Interval* interval = ndi.boundConfidence((*rules)[i]);
					double difference_of_bounds = interval->upper_bound - interval->lower_bound;
					delete interval;
					// There are few rare cases where the upper_bound will be _slightly_ less than the lower_bound due to rounding errors...
					if (difference_of_bounds < 0) {
						difference_of_bounds = 0;
					}
					if (difference_of_bounds == 0) {
						rules_with_zero_dob++;
					}
				
					if (difference_of_bounds > min_difference_of_bounds) {
						rules_output++;
						(*rules)[i]->store();
					} else {	
						// Pruned by dob
						delete (*rules)[i];
					}
				} else {
					// Pruned by confidence
					delete (*rules)[i];
				}
				rules_generated++;
			} else {
				(*rules)[i]->store();
			}
		}
		delete rules;
		
		for (int i = 0; i < iterator->children.size(); i++) {
			itemsets.push(iterator->children[i]);
		}
	}
	return rules_generated;
}

/**
 * Generates, counts and outputs rules from a frequent itemset tree to an output stream.
 * @param root The root of the frequent itemset tree.
 * @param os The output stream
 * @return The number of rules generated.
 */	
int RuleGenerator::generateRules(FITree* root, ostream& os) {
	queue<FITree*> itemsets;
	itemsets.push(root);

	FITree* iterator;
	vector<AssociationRule*>* rules;
	vector<AssociationRule*>* subrules;
	NDI ndi;

	while(!itemsets.empty()) {
		iterator = itemsets.front();
		itemsets.pop();
		
		rules = iterator->associationRules();
		for (int i = 0 ; i < rules->size(); i++) {
			// Only Non-Empty consequents
			if ((*rules)[i]->consequent->parent !=NULL) {
				double confidence = (double)((double)(*rules)[i]->itemset->support / (double)(*rules)[i]->condition->support);
				if (confidence >= min_confidence) {
					rules_with_high_confidence++;
				
					//Interval* interval = ndi.boundConfidenceCondition((*rules)[i]);
					//Interval* interval = ndi.boundConfidenceConsequent((*rules)[i]);
					Interval* interval = ndi.boundConfidence((*rules)[i]);
					double difference_of_bounds = interval->upper_bound - interval->lower_bound;
					delete interval;
					// There are few rare cases where the upper_bound will be _slightly_ less than the lower_bound due to rounding errors...
					if (difference_of_bounds < 0) {
						difference_of_bounds = 0;
					}
					if (difference_of_bounds == 0) {
						rules_with_zero_dob++;
					}
				
					if (difference_of_bounds > min_difference_of_bounds) {
						rules_output++;
						os << *((*rules)[i]) << "(" << confidence << ")" << " [" << interval->lower_bound << "," << interval->upper_bound << "] " << difference_of_bounds <<"\n";
						(*rules)[i]->store();
					} else {	
						// Pruned by dob
						delete (*rules)[i];
					}
				} else {
					// Pruned by confidence
					delete (*rules)[i];
				}
				rules_generated++;
			} else {
				(*rules)[i]->store();
			}
		}
		delete rules;
		
		for (int i = 0; i < iterator->children.size(); i++) {
			itemsets.push(iterator->children[i]);
		}
	}
	return rules_generated;
}

/**
 * Return the number of rules generated.
 * @return The total number of rules (pruned and non-pruned).
 */
int RuleGenerator::rulesGenerated() {
	return rules_generated;
}

/**
 * Return the number of rules output.
 * @return The number of output rules.
 */
int RuleGenerator::rulesOutput() {
	return rules_output;
}

/**
 * Return the number of confident rules.
 * @return The number of confident rules.
 */
int RuleGenerator::rulesConfident() {
	return rules_with_high_confidence;
}

/**
 * Return the number of rules with zero difference of bounds.
 * @return The number of rules with zero difference of bounds.
 */
int RuleGenerator::rulesWithZeroDob() {
	return rules_with_zero_dob;
}

/**
 * set the minimum difference of bounds
 * @param The minimum difference of bounds.
 */
void RuleGenerator::setMinDifferenceOfBounds(double d) {
	min_difference_of_bounds = d;
}

/**
 * Get the minimum difference of bounds
 * @return The minimum difference of bounds.
 */
double RuleGenerator::getMinDifferenceOfBounds() {
	return min_difference_of_bounds;
}

/**
 * Set the minimum confidence.
 * @param d The minimum confidence.
 */
void RuleGenerator::setMinConfidence(double d) {
	min_confidence = d;
}

/**
 * Get the minimum confidence
 * @return The minimum confidence.
 */
double RuleGenerator::getMinConfidence() {
	return min_confidence;
}
