/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
     
#include "SizeConstraint.h"
#include "FITree.h"
#include "AssociationRule.h"

/**
 * Create a size constraint with infinite maximum size.
 */
SizeConstraint::SizeConstraint() {
	last_size = 0;
	max_size = -1;
}

/**
 * Create a size constraint with a maximum size.
 * @param d The maximum size.
 */
SizeConstraint::SizeConstraint(int d) {
	last_size = 0;
	max_size = d;
}

/**
 * Set the maximum size of this constraint.
 * @param d The maximum size.
 */
void SizeConstraint::setMaxSize(int d) {
	max_size = d;
}

/**
 * Returns the result of last evaluation.
 * @return The size of the pattern evaluated last.
 */
double SizeConstraint::lastResult() {
	return last_size;
}

/**
 * Value an itemset.
 * @param fset The itemset.
 * @return The size of the itemset.
 */
double SizeConstraint::value(FITree* fset) {
	last_size = fset->size();
	return (double)last_size;
}

/**
 * Value an association rule.
 * @param arule The association rule.
 * @return The size of the association rule.
 */
double SizeConstraint::value(AssociationRule* arule) {
	last_size = arule->itemset->size();
	return (double)last_size;
}

/**
 * Validate an itemset.
 * @param fset The itemset.
 * @return True if the size of the itemset is higher than the maximum
 * size of this constraint.
 */
bool SizeConstraint::validate(FITree* fset) {
	if (max_size < 0) {
		return true;
	}
	
	if (value(fset) <= max_size) {
		return true;
	}
	
	return false;
}

/**
 * Validate an association rule.
 * @param arule The association rule.
 * @return True if the size of the rule is higher than the maximum
 * size of this constraint.
 */
bool SizeConstraint::validate(AssociationRule* arule) {
	if (max_size < 0) {
		return true;
	}

	if (value(arule) <= max_size) {
		return true;
	}
	
	return false;
}

/**
 * Returns the maximum size.
 * @return The maximum size.
 */
double SizeConstraint::constraintLimit() const {
	return max_size;
}

/**
 * Returns false.
 * @return false.
 */
bool SizeConstraint::increasing() const {
	return false;
}

/**
 * Returns 0.
 * @return 0
 */
int SizeConstraint::precision() const {
	return 0;
}

/**
 * Returns the name of this constraint.
 * @return The name.
 */
string SizeConstraint::name() const {
	return "Size";
}

/**
 * Returns the abbreviation of this constraint.
 * @return The abbreviation.
 */
string SizeConstraint::abbreviation() const {
	return "size";
}

/**
 * Returns the description of this constraint.
 * @return The description.
 */
string SizeConstraint::description() const {
	return "Size of the pattern is the number of items in the pattern.";
}
