/**
 * Copyright(c) Juho Muhonen 2004-2005
 * 
 * This file is part of 'firm'.
 * 
 * 'firm' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * 'firm' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with 'firm'; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
     
#include "TextOutputGenerator.h"
#include "SetConstraint.h"
#include "RuleConstraint.h"

#include "FITree.h"
#include "AssociationRule.h"

#include <queue>
using std::queue;

#include <iomanip>
#include <iostream>
using std::ostream;
using std::cout;
using std::endl;

/**
 * Creates a new text output generator.
 */
TextOutputGenerator::TextOutputGenerator() {
	output_stream = NULL;
	out = &cout;
	out->setf(std::ios::fixed);
	out->precision(3);
}

/**
 * The Destructor.
 */
TextOutputGenerator::~TextOutputGenerator() {
	finalize();
}

/**
 * Sets a new name for the output file. If there is an existing output file,
 * it is closed.
 * @param n The name of the outputfile.
 */
void TextOutputGenerator::setOutputName(string n) {
	assert(n != "");
	if (output_stream != NULL) {
		output_stream->close();
		delete output_stream;
	}	
	output_stream = new ofstream(n.c_str());
	out = output_stream;
	out->setf(std::ios::fixed);
	out->precision(3);
}

/**
 * Outputs the header for association rule output.
 */
void TextOutputGenerator::outputRuleMeasures() {
	*output_stream << "# Condition -> Consequent";
	for (vector<RuleConstraint*>::iterator cons_iter = rule_constraints.begin(); cons_iter != rule_constraints.end(); cons_iter++) { 
		if ((*cons_iter)->name() != "") {
			*output_stream << "|" << (*cons_iter)->name();
		}
	}
	*output_stream << endl;
}

/**
 * Outputs the header for itemset output.
 */
void TextOutputGenerator::outputItemsetMeasures() {
	*output_stream << "# Itemset";

	for (int i = 0; i < set_constraints.size(); i++) {
		*output_stream << "|" << set_constraints[i]->name();
	}

	*output_stream << endl;
}

/**
 * Adds a set constraint to this outputter.
 * @param c The set constraint.
 */
void TextOutputGenerator::addSetConstraint(SetConstraint* c) {
	assert(c != NULL);
	
	set_constraints.push_back(c);
}

/**
 * Adds a rule constraint to this outputter.
 * @param c The rule constraint.
 */
void TextOutputGenerator::addRuleConstraint(RuleConstraint* c) {
	assert(c != NULL);
	
	rule_constraints.push_back(c);
}

/**
 * Remove a constraint from this outputter.
 * @param s The name of the constraint.
 */
void TextOutputGenerator::removeConstraint(string s) {
	for (int i = 0; i < set_constraints.size(); i++) {
		if (set_constraints[i]->name() == s) {
			SetConstraint* s = set_constraints[i];
			set_constraints.erase(set_constraints.begin() + i);
			delete s;
		}
	}

	for (int i = 0; i < rule_constraints.size(); i++) {
		if (rule_constraints[i]->name() == s) {
			RuleConstraint* r = rule_constraints[i];
			rule_constraints.erase(rule_constraints.begin() + i);
			delete r;
		}
	}
}

/**
 * Output an itemsets.
 * @param set The frequent itemset.
 */
void TextOutputGenerator::outputSet(FITree* set) {
	assert(set != NULL);

	SetConstraint* constraint;

	bool valid = true;
	
	for (int i = 0; i < set_constraints.size() && valid; i++) {
		constraint = set_constraints[i];
		if (!constraint->validate(set)) {
			valid = false;
		}
	}

	if (valid) {
		*out << *set;
	
		for (int i = 0; i < set_constraints.size(); i++) {
			constraint = set_constraints[i];
			output_stream->precision(constraint->precision());
			//*out << "|" << constraint->abbreviation() << "=" << constraint->value(set);
			*out << "|" << constraint->value(set);
		}
	
		*out << endl;
	}
}


/**
 * Output all itemsets in a frequent itemset tree.
 * @param root The root of the frequent itemset tree.
 */
void TextOutputGenerator::outputSets(FITree* set) {
	assert(set != NULL);
	
	outputItemsetMeasures();
	
	outputSet(set);
	
	for (int i = 0; i < set->children.size(); i++) {
		outputSets(set->children[i]);
	}
}

/**
 * Output a single association rule.
 * @param rule The association rule.
 */
void TextOutputGenerator::outputRule(AssociationRule* rule) {
	assert(rule != NULL);

	RuleConstraint* constraint;
	bool valid = true;
	
	for (int i = 0; i < rule_constraints.size() && valid; i++) {
		constraint = rule_constraints[i];
		if (!constraint->validate(rule)) {
			valid = false;
		}
	}

	if (valid) {
		*out << *(rule->condition) << "-> " << *(rule->consequent) ;
	
		for (int i = 0; i < rule_constraints.size(); i++) {
			constraint = rule_constraints[i];
			output_stream->precision(constraint->precision());
			//*out << "|" << constraint->abbreviation() << "=" << constraint->value(rule);
			*out << "|" << constraint->value(rule);
		}
	
		*out << endl;
	}	
}

/**
 * Output all possible association rules from a set of itemsets.
 * @param root The root of the frequent itemset tree.
 */
void TextOutputGenerator::outputRules(FITree* root) {
	assert(root != NULL);

	outputRuleMeasures();

	queue<FITree*> itemsets;
	itemsets.push(root);

	FITree* iterator;
	vector<AssociationRule*>* rules;

	while(!itemsets.empty()) {
		iterator = itemsets.front();
		itemsets.pop();
		
		rules = iterator->associationRules();
		for (int i = 0 ; i < rules->size(); i++) {
			outputRule((*rules)[i]);
		}
		delete rules;
		
		for (int i = 0; i < iterator->children.size(); i++) {
			itemsets.push(iterator->children[i]);
		}
	}
}

/**
 * Finalizes this output generator after use. Namely flush and close output streams and delete all stored constraints.
 */
void TextOutputGenerator::finalize() {
	*out << std::flush;
	
	if (output_stream != NULL) {
		output_stream->close();
		delete output_stream;
	}

	for (int i = 0; i < set_constraints.size(); i++) {
		delete set_constraints[i];
	}

	for (int i = 0; i < rule_constraints.size(); i++) {
		delete rule_constraints[i];
	}
}
